#!/usr/bin/python3.7
########################################################################################
# __utils__.py - The utility module for the runs component of the pvt system model.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################

"""
The utility module for the PV-T system model's runs component.

This module contains common functionality, strucutres, and types, to be used by the
various modules throughout the PVT model.

"""

import datetime

from typing import Dict, List, Optional, Set, Union

import numpy

from .. import efficiency, index_handler
from ..pvt_collector import pvt

from ...__utils__ import OperatingMode, SystemData, TemperatureName
from ..__utils__ import (
    ProgrammerJudgementFault,
    WeatherConditions,
)
from ..constants import ZERO_CELCIUS_OFFSET
from ..physics_utils import reduced_temperature


__all__ = ("system_data_from_run",)


def _average_layer_temperature(
    pvt_collector: pvt.PVT,
    temperature_name: TemperatureName,
    temperature_vector: Union[List[float], numpy.ndarray],
) -> float:
    """
    Determines the average temperature for a layer.

    :param temperature_name:
        The name of the temperature for which to determine the average.

    :param temperature_vector:
        The temperature vector from which the average temperature of the given layer
        should be extracted.

    :return:
        The average temperature of the layer.

    """

    layer_temperatures: Set[float] = {
        value
        for value_index, value in enumerate(temperature_vector)
        if index_handler.temperature_name_from_index(value_index, pvt_collector)
        == temperature_name
    }

    try:
        average_temperature: float = sum(layer_temperatures) / len(layer_temperatures)
    except ZeroDivisionError as e:
        raise ProgrammerJudgementFault(
            "An attempt was made to compute the average temperature of a layer where "
            "no temperatures in the temperature vector matched the given temperatrure "
            f"name: {str(e)}"
        ) from None
    return average_temperature


def _layer_temperature_profile(
    number_of_pipes: int,
    number_of_x_elements: int,
    pvt_collector: pvt.PVT,
    temperature_name: TemperatureName,
    temperature_vector: Union[List[float], numpy.ndarray],
) -> Dict[str, float]:
    """
    Returns a map between coordinates and temperature in celcius for a layer.

    :param number_of_pipes:
        The number of pipes in the absorber.

    :param number_of_x_elements:
        The number of elements in a single row of the absorber.

    :param pvt_collector:
        The PVT panel being modelled.

    :param temperature_name:
        The name of the temperature for which to determine the average.

    :param temperature_vector:
        The temperature vector from which the average temperature of the given layer
        should be extracted.

    :return:
        A mapping between coordinates, expressed as a string, and the temperature of
        each element, for the layer specified. The values of the temperatures are
        expressed in Celcius.

    """

    if temperature_name in [
        TemperatureName.upper_glass,
        TemperatureName.glass,
        TemperatureName.pv,
        TemperatureName.absorber,
    ]:
        layer_temperature_map: Dict[str, float] = {
            str(element_coordinates): temperature_vector[
                index_handler.index_from_element_coordinates(
                    number_of_x_elements,
                    pvt_collector,
                    temperature_name,
                    element.x_index,
                    element.y_index,
                )
            ]
            - ZERO_CELCIUS_OFFSET
            for element_coordinates, element in pvt_collector.elements.items()
        }
    elif temperature_name in [
        TemperatureName.pipe,
        TemperatureName.htf,
        TemperatureName.htf_in,
        TemperatureName.htf_out,
    ]:
        layer_temperature_map = {
            str(element_coordinates): temperature_vector[
                index_handler.index_from_pipe_coordinates(
                    number_of_pipes,
                    number_of_x_elements,
                    element.pipe_index,  # type: ignore
                    pvt_collector,
                    temperature_name,
                    element.y_index,
                )
            ]
            - ZERO_CELCIUS_OFFSET
            for element_coordinates, element in pvt_collector.elements.items()
            if element.pipe
        }
    else:
        raise ProgrammerJudgementFault(
            "Attempt made to calculate a 2D profile for a temperature with no 2D "
            "nature."
        )

    return layer_temperature_map


def system_data_from_run(
    date: datetime.date,
    initial_date_and_time: datetime.datetime,
    number_of_pipes: int,
    number_of_x_elements: int,
    operating_mode: OperatingMode,
    pvt_collector: pvt.PVT,
    save_2d_output: bool,
    temperature_vector: Union[List[float], numpy.ndarray],
    time: datetime.time,
    weather_conditions: WeatherConditions,
) -> SystemData:
    """
    Return a :class:`SystemData` instance based on the current system.

    :param date:
        The date relevant to the system data being saved.

    :param initial_date_and_time:
        The initial date and time for the run.

    :param number_of_pies:
        The number of pipes attached to the absorber.

    :param number_of_x_elements:
        The number of x elements included in the absorber.

    :param operating_mode:
        The operating mode for the run.

    :param pvt_collector:
        A :class:`pvt.PVT` instance representing the PVT absorber being modelled.

    :param save_2d_output:
        Whether the 2D output should be saved (True) or not (False).

    :param temperature_vector:
        The temperature vector for which to compute the system variables.

    :param time:
        The time relevant to the system data being saved.

    :param weather_conditions:
        The weather conditions at the time being saved.

    :return:
        A :class:`..__utils__.SystemData` instance containing the information about the
        system at this point during the run.

    """

    # Determine the average temperatures of the various PVT layers.
    if any({element.upper_glass for element in pvt_collector.elements.values()}):
        average_upper_glass_temperature: Optional[float] = _average_layer_temperature(
            pvt_collector,
            TemperatureName.upper_glass,
            temperature_vector,
        )
        temperature_map_upper_glass_layer: Optional[
            Dict[str, float]
        ] = _layer_temperature_profile(
            number_of_pipes,
            number_of_x_elements,
            pvt_collector,
            TemperatureName.upper_glass,
            temperature_vector,
        )
    else:
        average_upper_glass_temperature = None
        temperature_map_upper_glass_layer = None

    if any({element.glass for element in pvt_collector.elements.values()}):
        average_glass_temperature: Optional[float] = _average_layer_temperature(
            pvt_collector,
            TemperatureName.glass,
            temperature_vector,
        )
        temperature_map_glass_layer: Optional[
            Dict[str, float]
        ] = _layer_temperature_profile(
            number_of_pipes,
            number_of_x_elements,
            pvt_collector,
            TemperatureName.glass,
            temperature_vector,
        )
    else:
        average_glass_temperature = None
        temperature_map_glass_layer = None

    average_pv_temperature = _average_layer_temperature(
        pvt_collector,
        TemperatureName.pv,
        temperature_vector,
    )
    temperature_map_pv_layer = _layer_temperature_profile(
        number_of_pipes,
        number_of_x_elements,
        pvt_collector,
        TemperatureName.pv,
        temperature_vector,
    )

    average_absorber_temperature = _average_layer_temperature(
        pvt_collector,
        TemperatureName.absorber,
        temperature_vector,
    )
    temperature_map_absorber_layer = _layer_temperature_profile(
        number_of_pipes,
        number_of_x_elements,
        pvt_collector,
        TemperatureName.absorber,
        temperature_vector,
    )

    average_pipe_temperature = _average_layer_temperature(
        pvt_collector,
        TemperatureName.pipe,
        temperature_vector,
    )
    temperature_map_pipe_layer = _layer_temperature_profile(
        number_of_pipes,
        number_of_x_elements,
        pvt_collector,
        TemperatureName.pipe,
        temperature_vector,
    )

    average_bulk_water_temperature = _average_layer_temperature(
        pvt_collector,
        TemperatureName.htf,
        temperature_vector,
    )
    temperature_map_bulk_water_layer = _layer_temperature_profile(
        number_of_pipes,
        number_of_x_elements,
        pvt_collector,
        TemperatureName.htf,
        temperature_vector,
    )

    # Set various parameters depending on the operating mode of the model.
    # Set the variables that depend on a coupled vs decoupled system.
    if operating_mode.coupled:
        exchanger_temperature_drop: Optional[float] = (
            temperature_vector[
                index_handler.index_from_temperature_name(
                    pvt_collector,
                    TemperatureName.tank_out,
                )
            ]
            - temperature_vector[
                index_handler.index_from_temperature_name(
                    pvt_collector,
                    TemperatureName.tank_in,
                )
            ]
            if temperature_vector[
                index_handler.index_from_temperature_name(
                    pvt_collector,
                    TemperatureName.tank_in,
                )
            ]
            > temperature_vector[
                index_handler.index_from_temperature_name(
                    pvt_collector,
                    TemperatureName.tank,
                )
            ]
            else 0
        )
        tank_temperature: Optional[float] = (
            temperature_vector[
                index_handler.index_from_temperature_name(
                    pvt_collector,
                    TemperatureName.tank,
                )
            ]
            - ZERO_CELCIUS_OFFSET
        )
    else:
        exchanger_temperature_drop = None
        tank_temperature = None
    # Set the variables that depend on a dynamic vs steady-state system.
    if operating_mode.dynamic:
        formatted_time: Optional[str] = (
            str((date.day - initial_date_and_time.day) * 24 + time.hour)
            + time.strftime("%H:%M:%S")[2:]
        )
    else:
        formatted_time = None

    # Compute variables in common to both.
    collector_input_temperature = temperature_vector[
        index_handler.index_from_temperature_name(
            pvt_collector,
            TemperatureName.collector_in,
        )
    ]
    collector_output_temperature = temperature_vector[
        index_handler.index_from_temperature_name(
            pvt_collector,
            TemperatureName.collector_out,
        )
    ]

    # Determine the reduced temperature of the system.
    if weather_conditions.irradiance > 0:
        electrical_efficiency: Optional[float] = efficiency.electrical_efficiency(
            pvt_collector, average_pv_temperature
        )
        reduced_system_temperature: Optional[float] = reduced_temperature(
            weather_conditions.ambient_temperature,
            average_bulk_water_temperature,
            weather_conditions.irradiance,
        )
        thermal_efficiency: Optional[float] = efficiency.thermal_efficiency(
            pvt_collector.area,
            pvt_collector.absorber.mass_flow_rate,
            weather_conditions.irradiance,
            collector_output_temperature - collector_input_temperature,
        )
    else:
        electrical_efficiency = None
        reduced_system_temperature = None
        thermal_efficiency = None

    # Return the system data.
    return SystemData(
        date=date.strftime("%d/%m/%Y"),
        time=formatted_time,
        upper_glass_temperature=average_upper_glass_temperature - ZERO_CELCIUS_OFFSET
        if average_upper_glass_temperature is not None
        else None,
        glass_temperature=average_glass_temperature - ZERO_CELCIUS_OFFSET
        if average_glass_temperature is not None
        else None,
        pv_temperature=average_pv_temperature - ZERO_CELCIUS_OFFSET,
        absorber_temperature=average_absorber_temperature - ZERO_CELCIUS_OFFSET,
        collector_input_temperature=collector_input_temperature - ZERO_CELCIUS_OFFSET,
        collector_output_temperature=collector_output_temperature - ZERO_CELCIUS_OFFSET,
        pipe_temperature=average_pipe_temperature - ZERO_CELCIUS_OFFSET,
        bulk_water_temperature=average_bulk_water_temperature - ZERO_CELCIUS_OFFSET,
        ambient_temperature=weather_conditions.ambient_temperature
        - ZERO_CELCIUS_OFFSET,
        exchanger_temperature_drop=exchanger_temperature_drop,
        sky_temperature=weather_conditions.sky_temperature - ZERO_CELCIUS_OFFSET,
        tank_temperature=tank_temperature,
        collector_temperature_gain=collector_output_temperature
        - collector_input_temperature,
        layer_temperature_map_bulk_water=temperature_map_bulk_water_layer
        if save_2d_output
        else None,
        layer_temperature_map_absorber=temperature_map_absorber_layer
        if save_2d_output
        else None,
        layer_temperature_map_glass=temperature_map_glass_layer
        if save_2d_output
        else None,
        layer_temperature_map_pipe=temperature_map_pipe_layer
        if save_2d_output
        else None,
        layer_temperature_map_pv=temperature_map_pv_layer if save_2d_output else None,
        layer_temperature_map_upper_glass=temperature_map_upper_glass_layer
        if save_2d_output
        else None,
        electrical_efficiency=electrical_efficiency,
        reduced_collector_temperature=reduced_system_temperature,
        thermal_efficiency=thermal_efficiency,
        solar_irradiance=weather_conditions.irradiance,
    )



#!/usr/bin/python3.7
########################################################################################
# constants.py - Module for holding and exporting fundamental constants.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################

"""
The Constants utility module for the pvt model component.

This module contains various physical constants and internal model-specific constants
which are accessed across multiple files.

"""

import datetime
import pytz

__all__ = (
    "ACCELERATION_DUE_TO_GRAVITY",
    "CONVERGENT_SOLUTION_PRECISION",
    "DEFAULT_INITIAL_DATE_AND_TIME",
    "DENSITY_OF_WATER",
    "EDGE_WIDTH",
    "EDGE_LENGTH",
    "HEAT_CAPACITY_OF_WATER",
    "HOT_WATER_DEMAND_TEMP",
    "INTERNAL_HOUSEHOLD_AMBIENT_TEMPERATURE",
    "MAXIMUM_RECURSION_DEPTH",
    "NUMBER_OF_COLLECTORS",
    "SPECIFIC_GAS_CONSTANT_OF_AIR",
    "STEFAN_BOLTZMAN_CONSTANT",
    "THERMAL_CONDUCTIVITY_OF_WATER",
    "WARN_RECURSION_DEPTH",
    "ZERO_CELCIUS_OFFSET",
)

#############
# Constants #
#############

# This scales the entire system's size - NOTE: Now depreciated.
NUMBER_OF_COLLECTORS = 1
# NUMBER_OF_COLLECTORS = 19414.28
# NUMBER_OF_COLLECTORS = 200

# The temperature of absolute zero in Kelvin, used for converting Celcius to Kelvin and
# vice-a-versa.
ZERO_CELCIUS_OFFSET: float = 273.15

# The acceleration due to gravity measured in meters per second squared.
ACCELERATION_DUE_TO_GRAVITY: float = 9.81

# The precision at which to calculate the convergent solution: 1 -> 0.1, 2 -> 0.01, etc.
CONVERGENT_SOLUTION_PRECISION = 1

# The initial date and time for the simultion to run from.
DEFAULT_INITIAL_DATE_AND_TIME = datetime.datetime(2005, 1, 1, 0, 0, tzinfo=pytz.UTC)

# The default system temperature, used for instantiating runs.
DEFAULT_SYSTEM_TEMPERATURE = ZERO_CELCIUS_OFFSET + 18.41  # [K]

# The density of water, measured in kilograms per meter cubed.
DENSITY_OF_WATER: int = 1000

# The maximum size allowed for the width of edge elements, measured in meters.
EDGE_WIDTH = 0.005

# The maximum size allowed for the length of top and bottom-eg
EDGE_LENGTH = 0.005

# The heat capacity of air, measured in Joules per kilogram Kelvin.
HEAT_CAPACITY_OF_AIR: int = 700

# The heat capacity of water, measured in Joules per kilogram Kelvin.
HEAT_CAPACITY_OF_WATER: int = 4182

# The initial temperature of the hot-water tank, at which it should be instantiated,
# measured in Kelvin.
# The temperature of hot-water required by the end-user, measured in Kelvin.
HOT_WATER_DEMAND_TEMP = 60 + ZERO_CELCIUS_OFFSET

# The average temperature of the air surrounding the tank, which is internal to the
# household, measured in Kelvin.
INTERNAL_HOUSEHOLD_AMBIENT_TEMPERATURE = ZERO_CELCIUS_OFFSET + 20  # [K]

# The maximum recursion depth for which to run the model. Beyond this, an error will be
# raised.
MAXIMUM_RECURSION_DEPTH = 20

# The gas constant of air in SI units of Joules per kilogram Kelvin.
SPECIFIC_GAS_CONSTANT_OF_AIR = 287.05  # [J/kg*K]

# The Stefan-Boltzman constant, given in Watts per meter squared Kelvin to the four.
STEFAN_BOLTZMAN_CONSTANT: float = 5.670374419 * (10 ** (-8))

# The thermal conductivity of water is obtained from
# http://hyperphysics.phy-astr.gsu.edu/hbase/Tables/thrcn.html
THERMAL_CONDUCTIVITY_OF_WATER: float = 0.5918  # [W/m*K]

# The recursion depth at which to raise a warning.
WARN_RECURSION_DEPTH = 10



#!/usr/bin/python3.7
########################################################################################
# pvt_collector/tedlar.py - Represents an tedlar layer within a PVT panel.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################

"""
The tedlar module for the PV-T model.

This module represents a tedlar layer within a PV-T panel.

"""

from .__utils__ import MicroLayer

__all__ = ("Tedlar",)


class Tedlar(MicroLayer):
    """
    Represents an Tedlar layer within the panel.

    """



#!/usr/bin/python3.7
########################################################################################
# pump.py - Represents a pump.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################

"""
Represents a pump.

"""

__all__ = ("Pump",)


class Pump:
    """
    Represents a physical pump in ths system.

    .. attribute:: on
        Whether the pump is currently on and pumping (True) or off (False). If the pump
        is off, there is no fluid flow through the pump.

    .. attribute:: power
        The electrical power consumed by the pump, measured in Watts.

    """

    def __init__(self, power: float) -> None:
        """
        Instantiates a :class:`Pump` instance.

        :param power:
            The power consumed by the pump when operating, measured in Watts.

        """

        self.on: bool = True
        self.power = power



#!/usr/bin/python3.7
########################################################################################
# process_pvt_system_data.py - Module for processing the PVT data files.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################

"""
The module for processing the PVT data files.

This module exposes various methods for instantiating and returning instances of the
classess representing physical components of the PVT system based on paths to various
data files.

Processing of these files, and extraction of their data to instantiate the components,
happens within this module.

"""

import datetime

from logging import Logger
from typing import Any, Dict, Optional, Set

from .pvt_collector import adhesive, bond, eva, pvt, element, tedlar
from . import exchanger, tank, pump

from ..__utils__ import (
    InvalidParametersError,
    MissingParametersError,
    read_yaml,
    TemperatureName,
)
from .constants import (
    EDGE_LENGTH,
    EDGE_WIDTH,
    HEAT_CAPACITY_OF_WATER,
)
from .__utils__ import (
    CollectorParameters,
    InvalidDataError,
    MissingDataError,
    OpticalLayerParameters,
    PVParameters,
)
from .index_handler import x_coordinate, y_coordinate
from .pvt_collector.__utils__ import MicroLayer


__all__ = (
    "heat_exchanger_from_path",
    "hot_water_tank_from_path",
    "pump_from_path",
    "pvt_collector_from_path",
)

#######################
# Heat exchanger code #
#######################


def heat_exchanger_from_path(exchanger_data_file: str) -> exchanger.Exchanger:
    """
    Generate a :class:`exchanger.Exchanger` instance based on the path to the data file.

    :param exchanger_data_file:
        The path to the exchanger data file.

    :return:
        A :class:`exchanger.Exchanger` instance representing the heat exchanger.

    """

    exchanger_data = read_yaml(exchanger_data_file)
    try:
        return exchanger.Exchanger(float(exchanger_data["efficiency"]))  # [unitless]
    except KeyError as e:
        raise MissingDataError(
            f"The file '{exchanger_data_file}' "
            f"is missing exchanger parameters: efficiency: {str(e)}"
        ) from None
    except ValueError as e:
        raise InvalidDataError(
            exchanger_data_file,
            "The exchanger efficiency must be a floating point integer.",
        ) from None


#######################
# Hot water tank code #
#######################


def hot_water_tank_from_path(tank_data_file: str) -> tank.Tank:
    """
    Generate a :class:`tank.Tank` instance based on the path to the data file.

    :param tank_data_file:
        The path to the tank data file.

    :param mains_water_temp:
        The mains-water temperature, measured in Kelvin.

    :return:
        A :class:`tank.Tank` instance representing the hot-water tank.

    """

    tank_data = read_yaml(tank_data_file)
    try:
        return tank.Tank(
            float(tank_data["area"]),  # [m^2]
            float(tank_data["diameter"]),  # [m]
            HEAT_CAPACITY_OF_WATER,  # [J/kg*K]
            float(tank_data["heat_loss_coefficient"]),  # [W/m^2*K]
            float(tank_data["mass"]),  # [kg]
        )
    except KeyError as e:
        raise MissingDataError(
            "Not all data needed to instantiate the tank class was provided. "
            f"File: {tank_data_file}. Error: {str(e)}"
        ) from None
    except ValueError as e:
        raise InvalidDataError(
            tank_data_file,
            "Tank data variables provided must be floating point integers.",
        ) from None


#############
# Pump code #
#############


def pump_from_path(pump_data_file: str) -> pump.Pump:
    """
    Generate a :class:`pump.Pump` instance based on the path to the data file.

    :param pump)data_file:
        The path to the pump data file.

    :return:
        A :class:`tank.Tank` instance representing the hot-water tank.

    """

    pump_data = read_yaml(pump_data_file)
    try:
        return pump.Pump(
            float(pump_data["power"]),  # [W]
        )
    except KeyError as e:
        raise MissingDataError(
            "Not all data needed to instantiate the pump class was provided. "
            f"File: {pump_data_file}. Error: {str(e)}"
        ) from None
    except ValueError as e:
        raise InvalidDataError(
            pump_data_file,
            "Pump data variables provided must be floating point integers.",
        ) from None


##################
# PVT panel code #
##################


def _absorber_params_from_data(
    length: float,
    absorber_data: Dict[str, Any],
    override_mass_flow_rate: Optional[float],
) -> CollectorParameters:
    """
    Generate a :class:`CollectorParameters` containing absorber-layer info from data.

    The HTF is assumed to be water unless the HTF heat capacity is supplied in the
    absorber data.

    :param length:
        The length of the PV-T system, measured in meters.

    :param absorber_data:
        The raw absorber data extracted from the YAML data file.

    :param override_mass_flow_rate:
        If specified, this overrides the mass flow-rate in the absorber. The units are
        litres per hour.

    :return:
        The absorber data, as a :class:`__utils__.CollectorParameters`, ready to
        instantiate a :class:`pvt.Collector` layer instance.

    """

    # import pdb

    # pdb.set_trace()

    try:
        return CollectorParameters(
            absorptivity=absorber_data["absorptivity"],  # [unitless]
            conductivity=absorber_data["thermal_conductivity"]
            if "thermal_conductivity" in absorber_data
            else None,
            density=absorber_data["density"] if "density" in absorber_data else None,
            emissivity=absorber_data["emissivity"],  # [unitless]
            heat_capacity=absorber_data["heat_capacity"],  # [J/kg*K]
            htf_heat_capacity=absorber_data["htf_heat_capacity"]  # [J/kg*K]
            if "htf_heat_capacity" in absorber_data
            else HEAT_CAPACITY_OF_WATER,  # [J/kg*K]
            inner_pipe_diameter=absorber_data["inner_pipe_diameter"],  # [m]
            length=length,  # [m]
            mass_flow_rate=override_mass_flow_rate
            if override_mass_flow_rate is not None
            else absorber_data["mass_flow_rate"],  # [Litres/hour]
            number_of_pipes=absorber_data["number_of_pipes"],  # [pipes]
            outer_pipe_diameter=absorber_data["outer_pipe_diameter"],  # [m]
            pipe_density=absorber_data["pipe_density"],  # [kg/m^3]
            thickness=absorber_data["thickness"],  # [m]
            transmissivity=absorber_data["transmissivity"],  # [unitless]
        )
    except KeyError as e:
        raise MissingDataError(
            "Not all needed absorber-layer data provided. Potential problem: absorber"
            "mass must be specified, either as 'mass' or 'density' and 'area' and "
            f"'thickness' params: {str(e)}"
        ) from None


def _glass_params_from_data(glass_data: Dict[str, Any]) -> OpticalLayerParameters:
    """
    Generate a :class:`OpticalLayerParameters` containing glass-layer info from data.

    :param area:
        The area of the PV-T system, measured in meters squared.

    :param glass_data:
        The raw glass data extracted from the YAML data file.

    :return:
        The glass data, as a :class:`__utils__.OpticalLayerParameters`, ready to
        instantiate a :class:`pvt.Glass` layer instance.

    """

    try:
        return OpticalLayerParameters(
            absorptivity=glass_data["absorptivity"],  # [unitless]
            conductivity=glass_data["thermal_conductivity"]
            if "thermal_conductivity" in glass_data
            else None,
            density=glass_data["density"],  # [kg/m^3]
            emissivity=glass_data["emissivity"],  # [unitless]
            heat_capacity=glass_data["heat_capacity"],  # [J/kg*K]
            thickness=glass_data["thickness"],  # [m]
            transmissivity=glass_data["transmissivity"],  # [unitless]
        )

    except KeyError as e:
        raise MissingDataError(
            "Not all needed glass-layer data provided. Potential problem: Glass mass "
            "must be specified, either as 'mass' or 'density' and 'area' and "
            f"'thickness' params: {str(e)}"
        ) from None


def _pv_params_from_data(pv_data: Optional[Dict[str, Any]]) -> PVParameters:
    """
    Generate a :class:`PVParameters` containing PV-layer info from data.

    :param area:
        The area of the PV-T system, measured in meters squared.

    :param pv_data:
        The raw PV data extracted from the YAML data file.

    :return:
        The PV data, as a :class:`__utils__.PVParameters`, ready to instantiate a
        :class:`pvt.PV` layer instance.

    """

    if pv_data is None:
        raise MissingParametersError(
            "PV", "PV parameters must be specified in the PVT YAML file."
        )

    try:
        return PVParameters(
            conductivity=pv_data["thermal_conductivity"]
            if "thermal_conductivity" in pv_data
            else None,
            density=pv_data["density"] if "density" in pv_data else None,  # [kg/m^3]
            heat_capacity=pv_data["heat_capacity"],  # [J/kg*K]
            thickness=pv_data["thickness"],  # [m]
            transmissivity=pv_data["transmissivity"],  # [unitless]
            absorptivity=pv_data["absorptivity"],  # [unitless]
            emissivity=pv_data["emissivity"],  # [unitless]
            reference_efficiency=pv_data["reference_efficiency"],  # [unitless]
            reference_temperature=pv_data["reference_temperature"],  # [K]
            thermal_coefficient=pv_data["thermal_coefficient"],  # [K^-1]
        )
    except KeyError as e:
        raise MissingDataError(
            "Not all needed PV-layer data provided. Potential problem: PV mass must be"
            "specified, either as 'mass' or 'density' and 'area' and 'thickness' "
            f"params: {str(e)}"
        ) from None


def _elements_from_data(
    edge_length: float,
    edge_width: float,
    layers: Set[TemperatureName],
    logger: Logger,
    portion_covered: float,
    pvt_data: Dict[Any, Any],
    x_resolution: int,
    y_resolution: int,
) -> Any:
    """
    Returns mapping from element coordinate to element based on the input data.

    :param edge_length:
        The maximum length of an edge element along the top and bottom edges of the
        panel, measured in meters.

    :param edge_width:
        The maximum width of an edge element along the side edges of the panel, measured
        in meters.

    :param layers:
        The `set` of layers to include in the system.

    :param logger:
        The :class:`logging.Logger` logger instance used for the run.

    :param portion_covered:
        The portion of the PVT absorber that is covered with PV cells. The uncovered
        section is mapped as solar absorber only with glazing as appropriate.

    :param pvt_data:
        The raw PVT data, extracted from the data file.

    :param x_resolution:
        The x resolution for the run.

    :param y_resolution:
        The y resolution for the run.

    :return:
        A mapping between the element coordinates and the element for all elements
        within the panel.

    """

    # * If 1x1, warn that 1x1 resolution is depreciated and should not really be used.
    if x_resolution == 1 and y_resolution == 1:
        logger.warn(
            "Running the system at a 1x1 resolution is depreciated. Consider running "
            "at a higher resolution."
        )
        return {
            element.ElementCoordinates(0, 0): element.Element(
                TemperatureName.absorber in layers,
                TemperatureName.glass in layers,
                pvt_data["pvt_collector"]["length"],
                True,
                TemperatureName.pv in layers,
                TemperatureName.upper_glass in layers,
                pvt_data["pvt_collector"]["width"],
                0,
                0,
                0,
            )
        }

    # Extract the necessary parameters from the system data.
    try:
        number_of_pipes = pvt_data["absorber"]["number_of_pipes"]
    except KeyError as e:
        raise MissingParametersError(
            "Element", "The number of pipes attached to the absorber must be supplied."
        ) from None
    try:
        panel_length = pvt_data["pvt_collector"]["length"]
    except KeyError as e:
        raise MissingParametersError(
            "Element", "PVT panel length must be supplied."
        ) from None

    try:
        panel_width = pvt_data["pvt_collector"]["width"]
    except KeyError as e:
        raise MissingParametersError(
            "Element", "PVT panel width must be supplied."
        ) from None
    try:
        bond_width = pvt_data["bond"]["width"]
    except KeyError as e:
        raise MissingParametersError(
            "Element", "Collector-to-pipe bond width must be supplied."
        ) from None

    # * Determine the spacing between the pipes.
    pipe_spacing = (x_resolution - number_of_pipes) / (number_of_pipes + 1)
    if int(pipe_spacing) != pipe_spacing:
        raise InvalidParametersError(
            "The resolution supplied results in an uneven pipe distribution.",
            "pipe_spcaing",
        )

    # * Determine the indicies of elements that have pipes attached.
    pipe_positions = list(
        range(int(pipe_spacing), x_resolution - 2, int(pipe_spacing) + 1)
    )

    # Determine whether the width of the elements is greater than or less than the edge
    # width and adjust accordingly.
    nominal_element_width: float = (
        panel_width - number_of_pipes * bond_width - 2 * edge_width
    ) / (x_resolution - number_of_pipes - 2)
    if nominal_element_width < edge_width:
        nominal_element_width = (panel_width - number_of_pipes * bond_width) / (
            x_resolution - number_of_pipes
        )
        edge_width = nominal_element_width

    # Likewise, determine whether the nominal element height is greater than the edge
    # height and adjust accordingly.
    nominal_element_length: float = (panel_length - 2 * edge_length) / (
        y_resolution - 2
    )
    if nominal_element_length < edge_length:
        nominal_element_length = panel_length / y_resolution
        edge_length = nominal_element_length

    # * Instantiate the array of elements.

    # Construct the elemented array based on the arguments.
    pv_coordinate_cutoff = int(y_resolution * portion_covered)
    try:
        elements = {
            element.ElementCoordinates(
                x_coordinate(element_number, x_resolution),
                y_coordinate(element_number, x_resolution),
            ): element.Element(
                absorber=TemperatureName.absorber in layers,
                glass=TemperatureName.glass in layers,
                length=edge_length
                if y_coordinate(element_number, x_resolution) in {0, y_resolution - 1}
                else nominal_element_length,
                pipe=x_coordinate(element_number, x_resolution) in pipe_positions
                if TemperatureName.pipe in layers
                else False,
                pv=y_coordinate(element_number, x_resolution) <= pv_coordinate_cutoff
                if TemperatureName.pv in layers
                else False,
                upper_glass=TemperatureName.upper_glass in layers,
                # Use the edge with if the element is an edge element.
                width=edge_width
                if x_coordinate(element_number, x_resolution) in {0, x_resolution - 1}
                # Otherwise, use the bond width if the element is a pipe element.
                else bond_width
                if x_coordinate(element_number, x_resolution) in pipe_positions
                # Otherwise, use the nominal element width.
                else nominal_element_width,
                x_index=x_coordinate(element_number, x_resolution),
                y_index=y_coordinate(element_number, x_resolution),
                pipe_index=pipe_positions.index(
                    x_coordinate(element_number, x_resolution)
                )
                if x_coordinate(element_number, x_resolution) in pipe_positions
                else None,
            )
            for element_number in range(x_resolution * y_resolution)
        }
    except KeyError as e:
        raise MissingParametersError(
            "PVT", f"Missing parameters when instantiating the PV-T system: {str(e)}"
        ) from None

    return elements


def pvt_collector_from_path(
    layers: Set[TemperatureName],
    logger: Logger,
    override_mass_flow_rate: Optional[float],
    portion_covered: float,
    pvt_data_file: str,
    x_resolution: int,
    y_resolution: int,
) -> pvt.PVT:
    """
    Generate a :class:`pvt.PVT` instance based on the path to the data file.

    :param layers:
        The `set` of layers that should be included in the panel.

    :param logger:
        The :class:`logging.Logger` used for the run.

    :param override_mass_flow_rate:
        If specified, the mass-flow rate in the collector uses this value rather than
        the value in the data file. The units are litres per hour.

    :param portion_covered:
        The portion of the PV-T panel which is covered in PV.

    :param pvt_data_file:
        The path to the pvt data file.

    :param x_resolution:
        The x resolution of the simulation being run.

    :param y_resolution:
        The y resolution of the simulation being run.

    :return:
        A :class:`pvt.PVT` instance representing the PVT panel.

    """

    # Parse the data file into the various data classes.
    pvt_data = read_yaml(pvt_data_file)

    # If there is upper-glass data present, process this accordingly.
    if "upper_glass" in pvt_data and TemperatureName.upper_glass in layers:
        upper_glass_parameters: Optional[
            OpticalLayerParameters
        ] = _glass_params_from_data(pvt_data["upper_glass"])
    elif "glass" in pvt_data and TemperatureName.upper_glass in layers:
        upper_glass_parameters = _glass_params_from_data(pvt_data["glass"])
    elif TemperatureName.upper_glass in layers:
        raise InvalidParametersError(
            "Upper glass layer requested on the command line but no glass data in pvt "
            "data file.",
            "glass",
        )
    else:
        upper_glass_parameters = None

    # If there is glass data present, process this accordingly.
    if "glass" in pvt_data and TemperatureName.glass in layers:
        air_gap_thickness = pvt_data["air_gap"]["thickness"]  # [m]
        glass_parameters: Optional[OpticalLayerParameters] = _glass_params_from_data(
            pvt_data["glass"]
        )
    elif TemperatureName.glass in layers:
        raise InvalidParametersError(
            "Glass layer requested on the command line but no glass data in pvt data "
            "file.",
            "glass",
        )
    else:
        air_gap_thickness = None
        glass_parameters = None

    pv_parameters = _pv_params_from_data(pvt_data["pv"] if "pv" in pvt_data else None)
    absorber_parameters = _absorber_params_from_data(
        pvt_data["pvt_collector"]["length"],  # [m]
        pvt_data["absorber"],
        override_mass_flow_rate,
    )

    elements = _elements_from_data(
        EDGE_LENGTH,
        EDGE_WIDTH,
        layers,
        logger,
        portion_covered,
        pvt_data,
        x_resolution,
        y_resolution,
    )

    try:
        pvt_collector = pvt.PVT(
            absorber_pipe_bond=bond.Bond(
                pvt_data["bond"]["thermal_conductivity"],
                pvt_data["bond"]["thickness"],
                pvt_data["bond"]["width"],
            ),
            adhesive=adhesive.Adhesive(
                pvt_data["adhesive"]["thermal_conductivity"],
                pvt_data["adhesive"]["thickness"],
            ),
            air_gap_thickness=air_gap_thickness,
            area=pvt_data["pvt_collector"]["area"]
            if "area" in pvt_data["pvt_collector"]
            else pvt_data["pvt_collector"]["width"]
            * pvt_data["pvt_collector"]["length"],  # [m^2]
            absorber_parameters=absorber_parameters,
            eva=eva.EVA(
                pvt_data["eva"]["thermal_conductivity"], pvt_data["eva"]["thickness"]
            ),
            glass_parameters=glass_parameters,
            insulation=MicroLayer(
                pvt_data["insulation"]["thermal_conductivity"],
                pvt_data["insulation"]["thickness"],
            ),
            latitude=pvt_data["pvt_collector"]["latitude"],  # [deg]
            length=pvt_data["pvt_collector"]["length"],  # [m]
            longitude=pvt_data["pvt_collector"]["longitude"],  # [deg]
            portion_covered=portion_covered,  # [unitless]
            pv_parameters=pv_parameters,
            elements=elements,
            tedlar=tedlar.Tedlar(
                pvt_data["tedlar"]["thermal_conductivity"],
                pvt_data["tedlar"]["thickness"],
            ),
            timezone=datetime.timezone(
                datetime.timedelta(hours=int(pvt_data["pvt_collector"]["timezone"]))
            ),
            upper_glass_parameters=upper_glass_parameters,
            width=pvt_data["pvt_collector"]["width"],  # [m]
            azimuthal_orientation=pvt_data["pvt_collector"][
                "azimuthal_orientation"
            ]  # [deg]
            if "azimuthal_orientation" in pvt_data["pvt_collector"]
            else None,
            horizontal_tracking=pvt_data["pvt_collector"]["horizontal_tracking"],
            vertical_tracking=pvt_data["pvt_collector"]["vertical_tracking"],
            tilt=pvt_data["pvt_collector"]["tilt"]  # [deg]
            if "tilt" in pvt_data["pvt_collector"]
            else None,
        )
    except KeyError as e:
        raise MissingParametersError(
            "PVT", f"Missing parameters when instantiating the PV-T system: {str(e)}"
        ) from None
    except TypeError as e:
        raise InvalidDataError(
            "PVT Data File", f"Error parsing data types - type mismatch: {str(e)}"
        ) from None

    return pvt_collector



#!/usr/bin/python3.7
########################################################################################
# continuity.py - The continuity module for the matrix component.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################

"""
The continuity module for the matrix component.

This module computes and returns the equation(s) associated with the continuity
equations of the PV-T collector for the matrix component.

The model works by arranging the system of differential equations as a matrix equation
such that the temperatures at each time step can be computed based on the coefficients
of the matrix which are computed based on the temperatures of the various components at
the previous time step, as well as various Physical and fundamental constants.

"""


from typing import List, Tuple, Union

import numpy

from .. import index_handler
from ..pvt_collector import pvt

from ...__utils__ import (
    TemperatureName,
)
from ..pvt_collector.element import Element

__all__ = (
    "calculate_decoupled_system_continuity_equation",
    "calculate_fluid_continuity_equation",
    "calculate_system_continuity_equations",
)


def calculate_decoupled_system_continuity_equation(
    collector_input_temperature: float,
    number_of_pipes: int,
    number_of_temperatures: int,
    number_of_x_elements: int,
    number_of_y_elements: int,
    pvt_collector: pvt.PVT,
) -> List[Tuple[List[float], float]]:
    """
    Compute the system continuity equations when the system is decoupled.

    The equations represented and computed are:
        - fluid enters the absorber at the set temperature specified (1);
        - the fluid entering the absorber is the same across all pipes (2);
        - fluid leaving the absorber is computed as an average over all the outputs
          (2);

    """

    equations: List[Tuple[List[float], float]] = list()

    # Equation 1: Fluid enters the absorber at the temperature specified.
    row_equation: List[float] = [0] * number_of_temperatures
    row_equation[
        index_handler.index_from_temperature_name(
            pvt_collector,
            TemperatureName.collector_in,
        )
    ] = 1
    equations.append((row_equation, collector_input_temperature))

    # Equation 2: Fluid entering the absorber is the same across all pipes.
    for pipe_number in range(number_of_pipes):
        row_equation = [0] * number_of_temperatures
        row_equation[
            index_handler.index_from_pipe_coordinates(
                number_of_pipes,
                number_of_x_elements,
                pipe_number,  # type: ignore
                pvt_collector,
                TemperatureName.htf_in,
                0,
            )
        ] = 1
        row_equation[
            index_handler.index_from_temperature_name(
                pvt_collector,
                TemperatureName.collector_in,
            )
        ] = -1
        equations.append((row_equation, 0))

    # Equation 3: Fluid leaving the absorber is computed by an average across the
    # output from all pipes.
    row_equation = [0] * number_of_temperatures
    for pipe_number in range(number_of_pipes):
        row_equation[
            index_handler.index_from_pipe_coordinates(
                number_of_pipes,
                number_of_x_elements,
                pipe_number,
                pvt_collector,
                TemperatureName.htf_out,
                number_of_y_elements - 1,
            )
        ] = (
            -1 / number_of_pipes
        )
    row_equation[
        index_handler.index_from_temperature_name(
            pvt_collector,
            TemperatureName.collector_out,
        )
    ] = 1
    equations.append((row_equation, 0))

    return equations


def calculate_fluid_continuity_equation(
    number_of_pipes: int,
    number_of_temperatures: int,
    number_of_x_elements: int,
    pvt_collector: pvt.PVT,
    element: Element,
) -> Tuple[List[float], float]:
    """
    Returns a matrix row and resultant vector value representing continuity of the htf.

    :param number_of_temperatures:
        The number of temperatures being modelled in the system.

    :return:
        A `tuple` containing:
        - the equation represented as a row in the matrix,
        - and the corresponding value in the resultant method.

    """

    # Compute the row equation
    row_equation: List[float] = [0] * number_of_temperatures

    # Compute the T_f,out(#, j) term.
    row_equation[
        index_handler.index_from_pipe_coordinates(
            number_of_pipes,
            number_of_x_elements,
            element.pipe_index,  # type: ignore
            pvt_collector,
            TemperatureName.htf_out,
            element.y_index,
        )
    ] = -1

    row_equation[
        index_handler.index_from_pipe_coordinates(
            number_of_pipes,
            number_of_x_elements,
            element.pipe_index,  # type: ignore
            pvt_collector,
            TemperatureName.htf_in,
            element.y_index + 1,
        )
    ] = 1

    return row_equation, 0


def calculate_system_continuity_equations(
    number_of_pipes: int,
    number_of_temperatures: int,
    number_of_x_elements: int,
    number_of_y_elements: int,
    previous_temperature_vector: Union[List[float], numpy.ndarray],
    pvt_collector: pvt.PVT,
) -> List[Tuple[List[float], float]]:
    """
    Returns matrix rows and resultant vector values representing system continuities.

    These inluce:
        - fluid entering the first section of the pipe is the same as that entering the
          absorber at the previous time step (1);
        - fluid leaving the absorber is an average over the various fluid temperatures
          leaving all pipes (2);
        - fluid entering the hot-water tank is the same as that leaving the absorber
          (3);
        - fluid leaving the hot-water tank is the same as that entering the absorber
          (4).

    :param number_of_temperatures:
        The number of temperatures being modelled in the system.

    :return:
        A `list` of `tuple`s containing:
        - the equation represented as a row in the matrix,
        - and the corresponding value in the resultant method.

    """

    equations: List[Tuple[List[float], float]] = list()

    # Equation 1: Continuity of fluid entering the absorber.
    for pipe_number in range(number_of_pipes):
        row_equation: List[float] = [0] * number_of_temperatures
        row_equation[
            index_handler.index_from_pipe_coordinates(
                number_of_pipes,
                number_of_x_elements,
                pipe_number,
                pvt_collector,
                TemperatureName.htf_in,
                0,
            )
        ] = 1
        resultant_value = previous_temperature_vector[
            index_handler.index_from_temperature_name(
                pvt_collector,
                TemperatureName.collector_in,
            )
        ]
        equations.append((row_equation, resultant_value))

    # Equation 2: Fluid leaving the absorber is computed by an average across the
    # output from all pipes.
    row_equation = [0] * number_of_temperatures
    for pipe_number in range(number_of_pipes):
        row_equation[
            index_handler.index_from_pipe_coordinates(
                number_of_pipes,
                number_of_x_elements,
                pipe_number,
                pvt_collector,
                TemperatureName.htf_out,
                number_of_y_elements - 1,
            )
        ] = (
            -1 / number_of_pipes
        )
    row_equation[
        index_handler.index_from_temperature_name(
            pvt_collector,
            TemperatureName.collector_out,
        )
    ] = 1
    equations.append((row_equation, 0))

    # Equation 3: Fluid leaving the absorber enters the tank without losses.
    row_equation = [0] * number_of_temperatures
    row_equation[
        index_handler.index_from_temperature_name(
            pvt_collector,
            TemperatureName.collector_out,
        )
    ] = -1
    row_equation[
        index_handler.index_from_temperature_name(
            pvt_collector,
            TemperatureName.tank_in,
        )
    ] = 1
    equations.append((row_equation, 0))

    row_equation = [0] * number_of_temperatures
    row_equation[
        index_handler.index_from_temperature_name(
            pvt_collector,
            TemperatureName.tank_out,
        )
    ] = -1
    row_equation[
        index_handler.index_from_temperature_name(
            pvt_collector,
            TemperatureName.collector_in,
        )
    ] = 1
    equations.append((row_equation, 0))

    return equations



#!/usr/bin/python3.7
# type: ignore
########################################################################################
# analysis.py - The analysis component for the model.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################
"""
Used for analysis of the output of the model runs.

NOTE: The mypy type checker is instructed to ignore this component. This is done due to
the lower standards applied to the analysis code, and the failure of mypy to correctly
type-check the external matplotlib.pyplot module.

"""

import argparse
import os
import sys

from logging import Logger
from typing import Any, List, Dict, Optional, Union

import yaml

from matplotlib import pyplot as plt

try:
    from ..__utils__ import get_logger
    from ..pvt_system.constants import (  # pylint: disable=unused-import
        HEAT_CAPACITY_OF_WATER,
    )
    from ..pvt_system.physics_utils import reduced_temperature
    from .__utils__ import (
        GraphDetail,
        load_model_data,
        plot_figure,
        plot_two_dimensional_figure,
    )
except ModuleNotFoundError:
    import logging

    logging.error(
        "Incorrect module import. Try running with `python3.7 -m pvt_model.analysis`"
    )
    raise

__all__ = ("analyse",)

# Used to distinguish copled data sets.
COUPLED_DATA_TYPE = "coupled"
# Used to distinguished decoupled data sets.
DECOUPLED_DATA_TYPE = "decoupled"
# Used to distinguish dynamic data sets.
DYNAMIC_DATA_TYPE = "dynamic"
# The directory into which which should be saved
NEW_FIGURES_DIRECTORY: str = "figures"
# Used to identify the "time" data set.
TIME_KEY = "time"
# The directory in which old figures are saved and stored for long-term access
OLD_FIGURES_DIRECTORY: str = "old_figures"
# Used to distinguish steady-state data sets.
STEADY_STATE_DATA_TYPE = "steady_state"
# Name of the steady-state data file.
STEADY_STATE_DATA_FILE_NAME = "autotherm.yaml"
# How detailed the graph should be
GRAPH_DETAIL: GraphDetail = GraphDetail.lowest
# How many values there should be between each tick on the x-axis
# X_TICK_SEPARATION: int = int(8 * GRAPH_DETAIL.value / 48)
X_TICK_SEPARATION: int = 8
# Which days of data to include
DAYS_TO_INCLUDE: List[bool] = [False, True]


def _parse_args(args) -> argparse.Namespace:
    """
    Parse the CLI args.

    """

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--data-file-name", "-df", help="Path to the data file to parse."
    )
    parser.add_argument(
        "--show-output",
        "-so",
        action="store_true",
        default=False,
        help="Show the output figures generated.",
    )
    parser.add_argument(
        "--skip-2d-plots",
        "-skip",
        action="store_true",
        default=False,
        help="Skip plotting of 2D figures.",
    )

    return parser.parse_args(args)


def _resolution_from_graph_detail(
    graph_detail: GraphDetail, num_data_points: int
) -> int:
    """
    Determine the x-step resolution of the plot.

    I realise here that the word "resolution" is over-used. Here, the number of data
    points (in the system data) that need to be absorbed into one graph point is
    calculated and returned.

    :param graph_detail:
        The detail level needed on the graph.

    :param num_data_points:
        The number of data points recorded in the plot.

    :return:
        The number of data points that need to be absorbed per graph point.

    """

    # * For "lowest", include one point every half-hour.
    if graph_detail == GraphDetail.highest:
        return 1

    return int(num_data_points / graph_detail.value)


def _reduce_data(  # pylint: disable=too-many-branches
    data_to_reduce: Dict[str, Dict[Any, Any]],
    graph_detail: GraphDetail,
    logger: Logger,
) -> Dict[Union[int, str], Dict[Any, Any]]:
    """
    This processes the data, using sums to reduce the resolution so it can be plotted.

    :param data_to_reduce:
        The raw, JSON data, contained within a `dict`.

    :param graph_detail:
        The level of detail required in the graph.

    :param logger:
        The logger being used in the run.

    :return:
        The cropped/summed up data, returned at a lower resolution as specified by the
        graph detail.

    """

    logger.info("Beginning data reduction.")

    # Determine the number of data points to be amalgamated per graph point.
    data_points_per_graph_point: int = _resolution_from_graph_detail(
        graph_detail, len(data_to_reduce)
    )

    logger.info(
        "%s data points will be condensed to each graph point.",
        data_points_per_graph_point,
    )

    if data_points_per_graph_point <= 1:
        return data_to_reduce

    # Construct a dictionary to contain this reduced data.
    reduced_data: Dict[Union[int, str], Dict[Any, Any]] = {
        index: dict()
        for index in range(
            int(
                len([key for key in data_to_reduce if key.isdigit()])
                / data_points_per_graph_point
            )
        )
    }

    # Depending on the type of data entry, i.e., whether it is a temperature, load,
    # demand covered, or irradiance (or other), the way that it is processed will vary.
    for data_entry_name in data_to_reduce[  # pylint: disable=too-many-nested-blocks
        "0"
    ].keys():
        # pdb.set_trace(header="Beginning of reduction loop.")
        # * If the entry is a date or time, just take the value
        if data_entry_name in ["date", TIME_KEY]:
            for index, _ in enumerate(reduced_data):
                reduced_data[index][data_entry_name] = data_to_reduce[
                    str(index * data_points_per_graph_point)
                ][data_entry_name]
            continue

        # * If the data entry is a temperature, or a power output, then take a rolling
        # average
        if any(
            (
                key in data_entry_name
                for key in [
                    "temperature",
                    "irradiance",
                    "efficiency",
                    "hot_water_load",
                    "electrical_load",
                ]
            )
        ):
            for outer_index, _ in enumerate(reduced_data):
                # Attempt to process as a dict or float first.
                if isinstance(data_to_reduce["0"][data_entry_name], float):
                    reduced_data[outer_index][data_entry_name] = sum(
                        [
                            float(data_to_reduce[str(inner_index)][data_entry_name])
                            / data_points_per_graph_point
                            for inner_index in range(
                                int(data_points_per_graph_point * outer_index),
                                int(data_points_per_graph_point * (outer_index + 1)),
                            )
                        ]
                    )
                elif isinstance(data_to_reduce["0"][data_entry_name], dict):
                    # Loop through the various coordinate pairs.
                    for sub_dict_key in data_to_reduce[str(outer_index)][
                        data_entry_name
                    ]:
                        if data_entry_name not in reduced_data[outer_index]:
                            reduced_data[outer_index][data_entry_name]: Dict[
                                str, float
                            ] = dict()
                        reduced_data[outer_index][data_entry_name][sub_dict_key] = sum(
                            [
                                float(
                                    data_to_reduce[str(inner_index)][data_entry_name][
                                        sub_dict_key
                                    ]
                                )
                                / data_points_per_graph_point
                                for inner_index in range(
                                    int(data_points_per_graph_point * outer_index),
                                    int(
                                        data_points_per_graph_point * (outer_index + 1)
                                    ),
                                )
                            ]
                        )
                else:
                    logger.debug(
                        "A value was an unsuported type. Setting to 'None': %s",
                        data_entry_name,
                    )
                    for index, _ in enumerate(reduced_data):
                        reduced_data[index][data_entry_name] = None
                    continue

        # * If the data entry is a load, then take a sum
        elif any((key in data_entry_name for key in {"load", "output"})):
            # @@@
            # * Here, the data is divided by 3600 to convert from Joules to Watt Hours.
            # * This only works provided that we are dealing with values in Joules...
            for outer_index, _ in enumerate(reduced_data):
                reduced_data[outer_index][data_entry_name] = (
                    sum(
                        [
                            float(data_to_reduce[str(inner_index)][data_entry_name])
                            for inner_index in range(
                                int(data_points_per_graph_point * outer_index),
                                int(data_points_per_graph_point * (outer_index + 1)),
                            )
                        ]
                    )
                    / 3600
                )
            continue

        # * Otherwise, we just use the data point.
        for index, _ in enumerate(reduced_data):
            reduced_data[index][data_entry_name] = data_to_reduce[
                str(index * data_points_per_graph_point)
            ][data_entry_name]
        continue

    return reduced_data


def _post_process_data(
    data_to_post_process: Dict[str, Dict[Any, Any]]
) -> Dict[str, Dict[Any, Any]]:
    """
    Carries out post-processing on data where necessary.

    Things to be computed:
        - Bulk Water Temperautre
        - Litres consumed

    :param data_to_post_process:
        The data to be post processed.

    :return:
        The post-processed data.

    """

    # * Cycle through all the data points and compute the new values as needed.
    for data_entry in data_to_post_process.values():
        data_entry["absorber_temperature_gain"] = (
            data_entry["collector_output_temperature"]
            - data_entry["collector_input_temperature"]
        )
    #     # Conversion needed from Wh to Joules.
    #     data_entry["litres_per_hour"] = (
    #         data_entry["thermal_load"] / (HEAT_CAPACITY_OF_WATER * 50) * 60
    #     )
    return data_to_post_process


def analyse_coupled_dynamic_data(
    data: Dict[Any, Any], logger: Logger, skip_2d_plots: bool
) -> None:
    """
    Carry out analysis on a set of dynamic data.

    :param data:
        The data to analyse.

    :param logger:
        The logger to use for the analysis run.

    :param skip_2d_plots:
        Whether to skip the 2D plots (True) or include them (False).

    """

    logger.info("Beginning analysis of coupled dynamic data set.")

    # * Reduce the resolution of the data.
    data = _reduce_data(data, GRAPH_DETAIL, logger)
    logger.info(
        "Data successfully reduced to %s graph detail level.", GRAPH_DETAIL.name
    )

    # * Create new data values where needed.
    data = _post_process_data(data)
    logger.info("Post-processing of data complete.")

    # Plot All Temperatures
    plot_figure(
        "all_temperatures",
        data,
        first_axis_things_to_plot=[
            "ambient_temperature",
            "bulk_water_temperature",
            "absorber_temperature",
            "collector_input_temperature",
            "collector_output_temperature",
            "glass_temperature",
            "pipe_temperature",
            "pv_temperature",
            "sky_temperature",
            "tank_temperature",
        ],
        first_axis_label="Temperature / deg C",
        first_axis_y_limits=[-10, 110],
    )

    # Plot All Temperatures
    plot_figure(
        "all_temperatures_unbounded",
        data,
        first_axis_things_to_plot=[
            "ambient_temperature",
            "bulk_water_temperature",
            "absorber_temperature",
            "collector_input_temperature",
            "collector_output_temperature",
            "glass_temperature",
            "pipe_temperature",
            "pv_temperature",
            "sky_temperature",
            "tank_temperature",
        ],
        first_axis_label="Temperature / deg C",
    )

    # # Plot Figure 4a: Electrical Demand
    # plot_figure(
    #     "maria_4a_electrical_load",
    #     data,
    #     ["electrical_load"],
    #     "Dwelling Load Profile / W",
    #     first_axis_y_limits=[0, 5000],
    #     first_axis_shape="d",
    # )

    # # Plot Figure 4b: Thermal Demand
    # plot_figure(
    #     "maria_4b_thermal_load",
    #     data,
    #     ["hot_water_load"],
    #     "Hot Water Consumption / Litres per hour",
    #     first_axis_y_limits=[0, 12],
    #     bar_plot=True,
    # )

    # # Plot Figure 5a: Diurnal Solar Irradiance
    # plot_figure(
    #     "maria_5a_solar_irradiance",
    #     data,
    #     [
    #         "solar_irradiance",
    #         # "normal_irradiance"
    #     ],
    #     "Solar Irradiance / Watts / meter squared",
    #     first_axis_y_limits=[0, 600],
    # )

    # Plot Figure 5b: Ambient Temperature
    plot_figure(
        "ambient_temperature",
        data,
        first_axis_things_to_plot=["ambient_temperature", "sky_temperature"],
        first_axis_label="Temperature / deg C",
        first_axis_y_limits=[0, 65],
    )

    # Plot Figure 6a: Panel-related Temperatures
    plot_figure(
        "panel_temperature",
        data,
        first_axis_things_to_plot=[
            "ambient_temperature",
            "bulk_water_temperature",
            "absorber_temperature",
            "glass_temperature",
            "pipe_temperature",
            "pv_temperature",
            "sky_temperature",
        ],
        first_axis_label="Temperature / deg C",
        first_axis_y_limits=[-10, 50],
    )

    # Plot Figure 6b: Tank-related Temperatures
    plot_figure(
        "tank_temperature",
        data,
        first_axis_things_to_plot=[
            "collector_output_temperature",
            "collector_input_temperature",
            "tank_temperature",
        ],
        first_axis_label="Temperature / deg C",
        first_axis_y_limits=[0, 50],
    )

    # Plot Figure 7: Stream-related Temperatures
    plot_figure(
        "stream_temperature",
        data,
        first_axis_things_to_plot=[
            "absorber_temperature_gain",
            "exchanger_temperature_drop",
        ],
        first_axis_label="Temperature Gain / deg C",
        # second_axis_things_to_plot=["tank_heat_addition"],
        # second_axis_label="Tank Heat Addition / W",
    )

    # # Plot Figure 8A - Electrical Power and Net Electrical Power
    # plot_figure(
    #     "maria_8a_electrical_output",
    #     data,
    #     ["gross_electrical_output", "net_electrical_output"],
    #     "Electrical Energy Supplied / Wh",
    # )

    # # Plot Figure 8B - Thermal Power Supplied and Thermal Power Demanded
    # plot_figure(
    #     "maria_8b_thermal_output",
    #     data,
    #     ["thermal_load", "thermal_output"],
    #     "Thermal Energy Supplied / Wh",
    # )

    # # Plot Figure 10 - Electrical Power, Gross only
    # plot_figure(
    #     "maria_10_gross_electrical_output",
    #     data,
    #     ["gross_electrical_output"],
    #     "Electrical Energy Supplied / Wh",
    # )

    # Plot glass layer temperatures at midnight, 6 am, noon, and 6 pm.
    if not skip_2d_plots:
        plot_two_dimensional_figure(
            "glass_temperature_0000",
            logger,
            data,
            "layer_temperature_map_glass",
            axis_label="Temperature / degC",
            hour=0,
            minute=0,
            plot_title="Glass layer temperature profile at 00:00",
        )

        plot_two_dimensional_figure(
            "glass_temperature_0600",
            logger,
            data,
            "layer_temperature_map_glass",
            axis_label="Temperature / degC",
            hour=6,
            minute=0,
            plot_title="Glass layer temperature profile at 06:00",
        )

        plot_two_dimensional_figure(
            "glass_temperature_1200",
            logger,
            data,
            "layer_temperature_map_glass",
            axis_label="Temperature / degC",
            hour=12,
            minute=0,
            plot_title="Glass layer temperature profile at 12:00",
        )

        plot_two_dimensional_figure(
            "glass_temperature_1800",
            logger,
            data,
            "layer_temperature_map_glass",
            axis_label="Temperature / degC",
            hour=18,
            minute=0,
            plot_title="Glass layer temperature profile at 18:00",
        )

        # Plot PV layer temperatures at midnight, 6 am, noon, and 6 pm.
        plot_two_dimensional_figure(
            "pv_temperature_0000",
            logger,
            data,
            "layer_temperature_map_pv",
            axis_label="Temperature / degC",
            hour=0,
            minute=0,
            plot_title="PV layer temperature profile at 00:00",
        )

        plot_two_dimensional_figure(
            "pv_temperature_0600",
            logger,
            data,
            "layer_temperature_map_pv",
            axis_label="Temperature / degC",
            hour=6,
            minute=0,
            plot_title="PV layer temperature profile at 06:00",
        )

        plot_two_dimensional_figure(
            "pv_temperature_1200",
            logger,
            data,
            "layer_temperature_map_pv",
            axis_label="Temperature / degC",
            hour=12,
            minute=0,
            plot_title="PV layer temperature profile at 12:00",
        )

        plot_two_dimensional_figure(
            "pv_temperature_1800",
            logger,
            data,
            "layer_temperature_map_pv",
            axis_label="Temperature / degC",
            hour=18,
            minute=0,
            plot_title="PV layer temperature profile at 18:00",
        )

        # Plot absorber layer temperatures at midnight, 6 am, noon, and 6 pm.
        plot_two_dimensional_figure(
            "absorber_temperature_0000",
            logger,
            data,
            "layer_temperature_map_absorber",
            axis_label="Temperature / degC",
            hour=0,
            minute=0,
            plot_title="Absorber layer temperature profile at 00:00",
        )

        plot_two_dimensional_figure(
            "absorber_temperature_0600",
            logger,
            data,
            "layer_temperature_map_absorber",
            axis_label="Temperature / degC",
            hour=6,
            minute=0,
            plot_title="Absorber layer temperature profile at 06:00",
        )

        plot_two_dimensional_figure(
            "absorber_temperature_1200",
            logger,
            data,
            "layer_temperature_map_absorber",
            axis_label="Temperature / degC",
            hour=12,
            minute=0,
            plot_title="Absorber layer temperature profile at 12:00",
        )

        plot_two_dimensional_figure(
            "absorber_temperature_1800",
            logger,
            data,
            "layer_temperature_map_absorber",
            axis_label="Temperature / degC",
            hour=18,
            minute=0,
            plot_title="Absorber layer temperature profile at 18:00",
        )

        # Plot bulk water temperatures at midnight, 6 am, noon, and 6 pm.
        plot_two_dimensional_figure(
            "pipe_temperature_0000",
            logger,
            data,
            "layer_temperature_map_pipe",
            axis_label="Temperature / degC",
            hour=0,
            minute=0,
            plot_title="Pipe temperature profile at 00:00",
        )

        plot_two_dimensional_figure(
            "pipe_temperature_0600",
            logger,
            data,
            "layer_temperature_map_pipe",
            axis_label="Temperature / degC",
            hour=6,
            minute=0,
            plot_title="Pipe temperature profile at 06:00",
        )

        plot_two_dimensional_figure(
            "pipe_temperature_1200",
            logger,
            data,
            "layer_temperature_map_pipe",
            axis_label="Temperature / degC",
            hour=12,
            minute=0,
            plot_title="Pipe temperature profile at 12:00",
        )

        plot_two_dimensional_figure(
            "pipe_temperature_1800",
            logger,
            data,
            "layer_temperature_map_pipe",
            axis_label="Temperature / degC",
            hour=18,
            minute=0,
            plot_title="Pipe temperature profile at 18:00",
        )

        # Plot bulk water temperatures at midnight, 6 am, noon, and 6 pm.
        plot_two_dimensional_figure(
            "bulk_water_temperature_0000",
            logger,
            data,
            "layer_temperature_map_bulk_water",
            axis_label="Temperature / degC",
            hour=0,
            minute=0,
            plot_title="Bulk-water temperature profile at 00:00",
        )

        plot_two_dimensional_figure(
            "bulk_water_temperature_0600",
            logger,
            data,
            "layer_temperature_map_bulk_water",
            axis_label="Temperature / degC",
            hour=6,
            minute=0,
            plot_title="Bulk-water temperature profile at 06:00",
        )

        plot_two_dimensional_figure(
            "bulk_water_temperature_1200",
            logger,
            data,
            "layer_temperature_map_bulk_water",
            axis_label="Temperature / degC",
            hour=12,
            minute=0,
            plot_title="Bulk-water temperature profile at 12:00",
        )

        plot_two_dimensional_figure(
            "bulk_water_temperature_1800",
            logger,
            data,
            "layer_temperature_map_bulk_water",
            axis_label="Temperature / degC",
            hour=18,
            minute=0,
            plot_title="Bulk-water temperature profile at 18:00",
        )

    """  # pylint: disable=pointless-string-statement
    # * Plotting all tank-related temperatures
    plot_figure(
        "tank_temperature",
        data,
        [
            "absorber_temperature",
            "collector_output_temperature",
            "absorber_temperature_gain",
            "tank_temperature",
            # "tank_output_temperature",
            "ambient_temperature",
            "sky_temperature",
        ],
        "Temperature / degC",
    )

    # * Plotting all temperatures relevant in the system.
    plot_figure(
        "all_temperatures",
        data,
        [
            "absorber_temperature",
            "collector_output_temperature",
            "absorber_temperature_gain",
            "tank_temperature",
            # "tank_output_temperature",
            "ambient_temperature",
            "sky_temperature",
            "glass_temperature",
            "pv_temperature",
        ],
        "Temperature / degC",
    )

    # * Plotting all PV-T panel layer temperatures
    plot_figure(
        "pvt_collector_temperature",
        data,
        [
            "glass_temperature",
            "pv_temperature",
            "absorber_temperature",
            "ambient_temperature",
            "sky_temperature",
        ],
        "Temperature / degC",
    )

    # * Plotting all temperatures in an unglazed panel
    plot_figure(
        "unglazed_pvt_temperature",
        data,
        [
            "pv_temperature",
            "absorber_temperature",
            "ambient_temperature",
            "sky_temperature",
        ],
        "Temperature / degC",
    )

    # * Plotting thermal-absorber-only temperatures
    plot_figure(
        "isolated_thermal_absorber",
        data,
        [
            "absorber_temperature",
            "ambient_temperature",
            "sky_temperature",
        ],
        "Temperature / degC",
    )

    # * Plotting demand covered and thermal load on one graph
    plot_figure(
        "demand_covered",
        data,
        first_axis_things_to_plot=["dc_electrical", "dc_thermal"],
        first_axis_label="Demand Covered / %",
        first_axis_y_limits=(0, 100),
        second_axis_things_to_plot=["thermal_load", "thermal_output"],
        second_axis_label="Thermal Energy Supplied / Wh",
    )

    # * Plotting the auxiliary heating required along with the thermal load on the
    # * system.

    plot_figure(
        "auxiliary_heating",
        data,
        first_axis_things_to_plot=["auxiliary_heating", "tank_heat_addition"],
        first_axis_label="Auxiliary Heating and Tank Heat Addition / Watts",
        second_axis_things_to_plot=["thermal_load", "thermal_output"],
        second_axis_label="Thermal Energy Supplied / Wh",
    )

    # * Plotting the absorber input, output, gain, and temperature.
    plot_figure(
        "absorber_temperatures",
        data,
        [
            "absorber_temperature",
            "collector_output_temperature",
            "collector_input_temperature",
            "absorber_temperature_gain",
            "tank_temperature",
        ],
        "Temperature / K",
    )

    # * Plotting the tank temperature, absorber temperature, and heat inputted into the
    # * tank.
    plot_figure(
        "tank_heat_gain_profile",
        data,
        first_axis_things_to_plot=["tank_temperature", "absorber_temperature"],
        first_axis_label="Temperature / deg C",
        first_axis_y_limits=(0, 100),
        second_axis_things_to_plot=["tank_heat_addition"],
        second_axis_label="Tank Heat Input / Watts",
    )

    # * Plotting the tank temperature, absorber temperature, and heat inputted into the
    # * tank.
    """  # pylint: disable=pointless-string-statement


def analyse_decoupled_steady_state_data(
    data: Dict[Any, Any], logger: Logger, skip_2d_plots: bool
) -> None:
    """
    Carry out analysis on a set of steady-state data.

    :param data:
        The data to analyse.

    :param logger:
        The logger to use for the analysis run.

    :param skip_2d_plots:
        Whether to skip the 2D plots (True) or plot them (False).

    """

    logger.info("Beginning steady-state analysis.")

    if skip_2d_plots:
        print(f"{int(len(data.keys()) * 2 + 2)} figures will be plotted.")
        logger.info("%s figures will be plotted.", int(len(data.keys()) * 2 + 2))
    else:
        print(f"{int(len(data.keys()) * 5 + 2)} figures will be plotted.")
        logger.info("%s figures will be plotted.", int(len(data.keys()) * 5 + 2))

    for temperature in data.keys():
        temperature_string = str(round(float(temperature), 2)).replace(".", "_")

        if not skip_2d_plots:
            # Glass Temperatures
            try:
                logger.info(
                    "Plotting 3D upper glass profile at %s degC.", temperature_string
                )
                plot_two_dimensional_figure(
                    "steady_state_upper_glass_layer_{}degC_input".format(
                        temperature_string
                    ),
                    logger,
                    data,
                    axis_label="Temperature / deg C",
                    entry_number=temperature,
                    plot_title="Upper glass layer temperature with {} K input HTF".format(
                        round(float(temperature), 2)
                    ),
                    thing_to_plot="layer_temperature_map_upper_glass",
                )
            except TypeError:
                logger.info(
                    "Upper-glass temperature profile could not be plotted due to no data."
                )

            # Glass Temperatures
            try:
                logger.info("Plotting 3D glass profile at %s degC.", temperature_string)
                plot_two_dimensional_figure(
                    "steady_state_glass_layer_{}degC_input".format(temperature_string),
                    logger,
                    data,
                    axis_label="Temperature / deg C",
                    entry_number=temperature,
                    plot_title="Glass layer temperature with {} K input HTF".format(
                        round(float(temperature), 2)
                    ),
                    thing_to_plot="layer_temperature_map_glass",
                )
            except TypeError:
                print("Glass temperature profile could not be plotted due to no data.")

            # PV Temperatures
            logger.info("Plotting 3D PV profile at %s degC.", temperature_string)
            plot_two_dimensional_figure(
                "steady_state_pv_layer_{}degC_input".format(temperature_string),
                logger,
                data,
                axis_label="Temperature / deg C",
                entry_number=temperature,
                plot_title="PV layer temperature with {} K input HTF".format(
                    round(float(temperature), 2)
                ),
                thing_to_plot="layer_temperature_map_pv",
            )

            # Collector Temperatures
            logger.info("Plotting 3D absorber profile at %s degC.", temperature_string)
            plot_two_dimensional_figure(
                "steady_state_absorber_layer_{}degC_input".format(temperature_string),
                logger,
                data,
                axis_label="Temperature / deg C",
                entry_number=temperature,
                plot_title="Collector layer temperature with {} K input HTF".format(
                    round(float(temperature), 2)
                ),
                thing_to_plot="layer_temperature_map_absorber",
            )

        # Pipe Temperatures
        logger.info(
            "Plotting 3D pipe profile at %s degC. NOTE: The profile will appear 2D if "
            "only one pipe is present.",
            temperature_string,
        )
        plot_two_dimensional_figure(
            "steady_state_pipe_{}degC_input".format(temperature_string),
            logger,
            data,
            axis_label="Pipe temperature / deg C",
            entry_number=temperature,
            plot_title="Pipe temperature with {} K input HTF".format(
                round(float(temperature), 2)
            ),
            thing_to_plot="layer_temperature_map_pipe",
        )

        # Bulk-water Temperatures
        logger.info(
            "Plotting 3D bulk-water profile at %s degC. NOTE: The profile will appear "
            "2D if only one pipe is present.",
            temperature_string,
        )
        plot_two_dimensional_figure(
            "steady_state_bulk_water_{}degC_input".format(temperature_string),
            logger,
            data,
            axis_label="Bulk-water temperature / deg C",
            entry_number=temperature,
            plot_title="Bulk-water temperature with {} K input HTF".format(
                round(float(temperature), 2)
            ),
            thing_to_plot="layer_temperature_map_bulk_water",
        )

    # Parse the thermal-efficiency data.
    with open(
        os.path.join("system_data", "steady_state_data", STEADY_STATE_DATA_FILE_NAME),
        "r",
    ) as f:  #
        experimental_steady_state_data = yaml.safe_load(f)

    # Post-process this data.
    for entry in experimental_steady_state_data:
        entry["reduced_temperature"] = reduced_temperature(
            entry["ambient_temperature"],
            entry["average_bulk_water_temperature"],
            entry["irradiance"],
        )

    # Thermal efficiency plot.
    logger.info("Plotting thermal efficiency against the reduced temperature.")

    # Plot the experimental data.
    _, ax1 = plt.subplots()
    ax1.scatter(
        [entry["reduced_temperature"] for entry in experimental_steady_state_data],
        [entry["thermal_efficiency"] for entry in experimental_steady_state_data],
        marker="s",
    )

    # Add the model data.
    plot_figure(
        "thermal_efficiency_against_reduced_temperature",
        data,
        first_axis_things_to_plot=["thermal_efficiency"],
        first_axis_label="Thermal efficiency",
        x_axis_label="Reduced temperature / K m^2 / W",
        x_axis_thing_to_plot="reduced_collector_temperature",
        plot_title="Thermal efficiency against reduced temperature",
        disable_lines=True,
        override_axis=ax1,
    )

    # Collector temperature gain plot.
    logger.info(
        "Plotting collector temperature gain against the input HTF temperature."
    )

    # Plot the experimental data.
    _, ax1 = plt.subplots()
    ax1.scatter(
        [
            entry["collector_input_temperature"]
            for entry in experimental_steady_state_data
        ],
        [
            entry["collector_temperature_gain"]
            for entry in experimental_steady_state_data
        ],
        marker="s",
    )

    # Add the model data.
    plot_figure(
        "collector_tempreature_gain_against_input_temperature",
        data,
        first_axis_things_to_plot=["collector_temperature_gain"],
        first_axis_label="Collector temperature gain / K",
        x_axis_label="Collector input temperature / degC",
        use_data_keys_as_x_axis=True,
        plot_title="Collector temperature gain against input temperature",
        disable_lines=True,
        override_axis=ax1,
    )

    # Plot the electrical efficiency against the reduced temperature.
    plot_figure(
        "electrical_efficiency_against_reduced_temperature",
        data,
        first_axis_things_to_plot=["electrical_efficiency"],
        first_axis_label="Electrical efficiency",
        x_axis_label="Reduced temperature / K m^2 / W",
        x_axis_thing_to_plot="reduced_collector_temperature",
        plot_title="Electrical efficiency against reduced temperature",
        disable_lines=True,
    )


def analyse_decoupled_dynamic_data(data: Dict[Any, Any], logger: Logger) -> None:
    """
    Carry out analysis on a decoupled dyanmic set of data.

    :param data:
        The data to analyse.

    :param logger:
        The logger to use for the run.

    """

    logger.info("Beginning analysis of a decoupled dynamic data set.")

    # * Reduce the resolution of the data.
    data = _reduce_data(data, GRAPH_DETAIL, logger)
    logger.info(
        "Data successfully reduced to %s graph detail level.", GRAPH_DETAIL.name
    )

    # * Create new data values where needed.
    data = _post_process_data(data)
    logger.info("Post-processing of data complete.")

    # Clip out the data points up to 10 minutes in.
    data = {key: value for key, value in data.items() if int(key) >= 30}

    # Plot output temperature and irradiance.
    plot_figure(
        "collector_output_response",
        data,
        first_axis_things_to_plot=[
            "collector_output_temperature",
            "collector_input_temperature",
            "ambient_temperature",
        ],
        first_axis_label="Collector Output Temperature / deg C",
        first_axis_y_limits=[15, 25],
        second_axis_things_to_plot=[
            "solar_irradiance",
        ],
        second_axis_label="Solar Irradiance / W/m^2",
        second_axis_y_limits=[0, 1000],
    )


def analyse(
    data_file_name: str,
    show_output: Optional[bool] = False,
    skip_2d_plots: Optional[bool] = False,
) -> None:
    """
    The main method for the analysis module.

    :param data_file_name:
        The path to the data file to analyse.

    :param show_output:
        Whether to show the output files generated.

    :param skip_2d_plots:
        Whether to skip the 2D plots (True) of include them (False).

    """

    # * Set up the logger
    logger = get_logger(True, "pvt_analysis", True)

    # * Extract the data.
    data = load_model_data(data_file_name)

    # * Determine whether the data is dynamic or steady-state.
    try:
        data_type = data.pop("data_type")
    except KeyError:
        logger.error(
            "Analysis data without an explicit data type is depreciated. Dynamic assumed."
        )
        data_type = DYNAMIC_DATA_TYPE

    # * Carry out analysis appropriate to the data type specified.
    if data_type in (DYNAMIC_DATA_TYPE, f"{COUPLED_DATA_TYPE}_{DYNAMIC_DATA_TYPE}"):
        analyse_coupled_dynamic_data(data, logger, skip_2d_plots)
    elif data_type in (
        STEADY_STATE_DATA_TYPE,
        f"{DECOUPLED_DATA_TYPE}_{STEADY_STATE_DATA_TYPE}",
    ):
        analyse_decoupled_steady_state_data(data, logger, skip_2d_plots)
    elif data_type == f"{DECOUPLED_DATA_TYPE}_{DYNAMIC_DATA_TYPE}":
        analyse_decoupled_dynamic_data(data, logger)
    else:
        logger.error("Data type was neither 'dynamic' nor 'steady_state'. Exiting...")
        sys.exit(1)

    logger.info("Analysis complete - all figures saved successfully.")

    if show_output:
        plt.show()


if __name__ == "__main__":
    parsed_args = _parse_args(sys.argv[1:])
    analyse(
        parsed_args.data_file_name, parsed_args.show_output, parsed_args.skip_2d_plots
    )



#!/usr/bin/python3.7
# type: ignore
########################################################################################
# glazing_analysis.py - The glazing analysis component for the model.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################
"""
Used for analysis of various differences between outputs from model runs.

NOTE: The mypy type checker is instructed to ignore this component. This is done due to
the lower standards applied to the analysis code, and the failure of mypy to correctly
type-check the external matplotlib.pyplot module.

"""

import argparse
import collections
import os
import sys

from logging import Logger
from typing import Any, List, Dict, Optional, Set

import re

from matplotlib import pyplot as plt

try:
    from ..__utils__ import get_logger
    from ..pvt_system.constants import (  # pylint: disable=unused-import
        HEAT_CAPACITY_OF_WATER,
    )
    from .__utils__ import (
        GraphDetail,
        load_model_data,
        plot_figure,
    )
except ModuleNotFoundError:
    import logging

    logging.error(
        "Incorrect module import. Try running with `python3.7 -m pvt_model.analysis`"
    )
    raise

__all__ = ("analyse",)

# Used to distinguish copled data sets.
COUPLED_DATA_TYPE = "coupled"
# Used to distinguished decoupled data sets.
DECOUPLED_DATA_TYPE = "decoupled"
# Used to distinguish dynamic data sets.
DYNAMIC_DATA_TYPE = "dynamic"
# The directory into which which should be saved
NEW_FIGURES_DIRECTORY: str = "figures"
# Used to identify the "time" data set.
TIME_KEY = "time"
# The directory in which old figures are saved and stored for long-term access
OLD_FIGURES_DIRECTORY: str = "old_figures"
# Used to distinguish steady-state data sets.
STEADY_STATE_DATA_TYPE = "steady_state"
# Name of the steady-state data file.
STEADY_STATE_DATA_FILE_NAME = "autotherm.yaml"
# How detailed the graph should be
GRAPH_DETAIL: GraphDetail = GraphDetail.lowest
# How many values there should be between each tick on the x-axis
# X_TICK_SEPARATION: int = int(8 * GRAPH_DETAIL.value / 48)
X_TICK_SEPARATION: int = 8
# Which days of data to include
DAYS_TO_INCLUDE: List[bool] = [False, True]
# Portion-covered regex.
PORTION_COVERED_REGEX = re.compile(
    r".*pc_(?P<first_digit>[0-9]*)_(?P<second_digit>[0-9])_.*"
)


def _parse_args(args) -> argparse.Namespace:
    """
    Parse the CLI args.

    """

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--data-file-directory",
        "-dfdir",
        help="Path to the directory containing data files to parse.",
    )
    parser.add_argument(
        "--show-output",
        "-so",
        action="store_true",
        default=False,
        help="Show the output figures generated.",
    )

    return parser.parse_args(args)


def _resolution_from_graph_detail(
    graph_detail: GraphDetail, num_data_points: int
) -> int:
    """
    Determine the x-step resolution of the plot.

    I realise here that the word "resolution" is over-used. Here, the number of data
    points (in the system data) that need to be absorbed into one graph point is
    calculated and returned.

    :param graph_detail:
        The detail level needed on the graph.

    :param num_data_points:
        The number of data points recorded in the plot.

    :return:
        The number of data points that need to be absorbed per graph point.

    """

    # * For "lowest", include one point every half-hour.
    if graph_detail == GraphDetail.highest:
        return 1

    return int(num_data_points / graph_detail.value)


def _post_process_data(
    data_to_post_process: Dict[str, Dict[Any, Any]]
) -> Dict[str, Dict[Any, Any]]:
    """
    Carries out post-processing on data where necessary.

    Things to be computed:
        - Bulk Water Temperautre
        - Litres consumed

    :param data_to_post_process:
        The data to be post processed.

    :return:
        The post-processed data.

    """

    # * Cycle through all the data points and compute the new values as needed.
    for data_entry in data_to_post_process.values():
        data_entry["absorber_temperature_gain"] = (
            data_entry["collector_output_temperature"]
            - data_entry["collector_input_temperature"]
        )
    return data_to_post_process


def _efficiency_plots(
    *,
    data: Dict[str, Any],
    electrical_efficiency_labels: List[str],
    logger: Logger,
    prefix: str,
    thermal_efficiency_labels: List[str],
) -> None:
    """
    Plot, based on the data provided, the thermal and electrical efficiencies.

    :param data:
        The data to be plotted of the type from each data file.

    :param electrical_efficiency_labels:
        A `list` containing the names assigned to the various electrical efficiency
        data sets stored in the dict entries.

    :param logger:
        The logger to use for the run.

    :param prefix:
        A prefix to put at the start of the file names.

    :param thermal_efficiency_labels:
        A `list` containing the names assigned to the various thermal efficiency data
        sets stored in the dict entries.

    """

    # Thermal efficiency plot.
    logger.info("Plotting thermal efficiency against the reduced temperature.")
    plot_figure(
        f"{prefix}_thermal_efficiency_against_reduced_temperature",
        data,
        first_axis_things_to_plot=thermal_efficiency_labels,
        first_axis_label="Thermal efficiency",
        x_axis_label="Reduced temperature / K m^2 / W",
        x_axis_thing_to_plot="reduced_collector_temperature",
        plot_title="Thermal efficiency against reduced temperature",
        disable_lines=True,
        plot_trendline=True,
    )

    # Collector temperature gain plot.
    # logger.info(
    #     "Plotting collector temperature gain against the input HTF temperature."
    # )

    # plot_figure(
    #     f"{prefix}_collector_tempreature_gain_against_input_temperature",
    #     data,
    #     first_axis_things_to_plot=["collector_temperature_gain"],
    #     first_axis_label="Collector temperature gain / K",
    #     x_axis_label="Collector input temperature / degC",
    #     use_data_keys_as_x_axis=True,
    #     plot_title="Collector temperature gain against input temperature",
    #     disable_lines=True,
    # )

    # Plot the electrical efficiency against the reduced temperature.
    logger.info("Plotting electrical efficiency against the reduced temperature.")

    plot_figure(
        f"{prefix}_electrical_efficiency_against_reduced_temperature",
        data,
        first_axis_things_to_plot=electrical_efficiency_labels,
        first_axis_label="Electrical efficiency",
        x_axis_label="Reduced temperature / K m^2 / W",
        x_axis_thing_to_plot="reduced_collector_temperature",
        plot_title="Electrical efficiency against reduced temperature",
        disable_lines=True,
        plot_trendline=True,
        first_axis_y_limits=[0.08, 0.14],
    )


def analyse_decoupled_steady_state_data(  # pylint: disable=too-many-branches
    data: Dict[Any, Dict[str, Any]], logger: Logger
) -> None:
    """
    Carry out analysis on a series of of steady-state data sets.

    :param data:
        The data to analyse. This should be a `dict` mapping the file name to the data
        extracted from the file.

    :param logger:
        The logger to use for the analysis run.

    """

    logger.info("Beginning steady-state analysis.")

    # Plot the portion-covered affect on an unglazed Ilaria.
    reduced_data: Dict[str, Any] = collections.defaultdict(dict)
    thermal_efficiency_labels: Set[str] = set()
    electrical_efficiency_labels: Set[str] = set()
    for key, sub_dict in data.items():
        if "ilaria_unglazed_pc_" not in key:
            continue
        portion_covered_match = re.match(PORTION_COVERED_REGEX, key)
        if portion_covered_match is None:
            sys.exit(1)
        portion_covered = float(
            f"{portion_covered_match.group('first_digit')}.{portion_covered_match.group('second_digit')}"
        )
        # Only plot significant portions covered.
        if portion_covered not in {1.0, 0.6, 0.3}:
            continue
        for sub_key, value in sub_dict.items():
            # Fetch the portion covered.
            # Store the specific thermal efficiency.
            reduced_data[sub_key][f"thermal efficiency p.c.={portion_covered}"] = value[
                "thermal_efficiency"
            ]
            thermal_efficiency_labels.add(f"thermal efficiency p.c.={portion_covered}")
            # Store the specific electrical efficiency.
            reduced_data[sub_key][
                f"electrical efficiency p.c.={portion_covered}"
            ] = value["electrical_efficiency"]
            electrical_efficiency_labels.add(
                f"electrical efficiency p.c.={portion_covered}"
            )
            # Store a copy of the reduced temperature
            reduced_data[sub_key]["reduced_collector_temperature"] = value[
                "reduced_collector_temperature"
            ]

    _efficiency_plots(
        data=reduced_data,
        electrical_efficiency_labels=sorted(electrical_efficiency_labels),
        logger=logger,
        prefix="ilaria_unglazed_pc",
        thermal_efficiency_labels=sorted(thermal_efficiency_labels),
    )
    plt.close("all")

    # Plot the portion-covered affect on unglazed autotherm.
    reduced_data = collections.defaultdict(dict)
    thermal_efficiency_labels = set()
    electrical_efficiency_labels = set()
    for key, sub_dict in data.items():
        if "autotherm_unglazed_pc_" not in key:
            continue
        portion_covered_match = re.match(PORTION_COVERED_REGEX, key)
        if portion_covered_match is None:
            sys.exit(1)
        portion_covered = float(
            f"{portion_covered_match.group('first_digit')}.{portion_covered_match.group('second_digit')}"
        )
        # Only plot significant portions covered.
        if portion_covered not in {1.0, 0.6, 0.3}:
            continue
        for sub_key, value in sub_dict.items():
            # Fetch the portion covered.
            # Store the specific thermal efficiency.
            reduced_data[sub_key][f"thermal efficiency p.c.={portion_covered}"] = value[
                "thermal_efficiency"
            ]
            thermal_efficiency_labels.add(f"thermal efficiency p.c.={portion_covered}")
            # Store the specific electrical efficiency.
            reduced_data[sub_key][
                f"electrical efficiency p.c.={portion_covered}"
            ] = value["electrical_efficiency"]
            electrical_efficiency_labels.add(
                f"electrical efficiency p.c.={portion_covered}"
            )
            # Store a copy of the reduced temperature
            reduced_data[sub_key]["reduced_collector_temperature"] = value[
                "reduced_collector_temperature"
            ]

    _efficiency_plots(
        data=reduced_data,
        electrical_efficiency_labels=sorted(electrical_efficiency_labels),
        logger=logger,
        prefix="autotherm_unglazed_pc",
        thermal_efficiency_labels=sorted(thermal_efficiency_labels),
    )
    plt.close("all")

    # Plot the portion-covered affect on single-glazed Ilaria.
    reduced_data = collections.defaultdict(dict)
    thermal_efficiency_labels = set()
    electrical_efficiency_labels = set()
    for key, sub_dict in data.items():
        if "ilaria_single_glazed_pc_" not in key:
            continue
        portion_covered_match = re.match(PORTION_COVERED_REGEX, key)
        if portion_covered_match is None:
            sys.exit(1)
        portion_covered = float(
            f"{portion_covered_match.group('first_digit')}.{portion_covered_match.group('second_digit')}"
        )
        # Only plot significant portions covered.
        if portion_covered not in {1.0, 0.6, 0.3}:
            continue
        for sub_key, value in sub_dict.items():
            # Fetch the portion covered.
            # Store the specific thermal efficiency.
            reduced_data[sub_key][f"thermal efficiency p.c.={portion_covered}"] = value[
                "thermal_efficiency"
            ]
            thermal_efficiency_labels.add(f"thermal efficiency p.c.={portion_covered}")
            # Store the specific electrical efficiency.
            reduced_data[sub_key][
                f"electrical efficiency p.c.={portion_covered}"
            ] = value["electrical_efficiency"]
            electrical_efficiency_labels.add(
                f"electrical efficiency p.c.={portion_covered}"
            )
            # Store a copy of the reduced temperature
            reduced_data[sub_key]["reduced_collector_temperature"] = value[
                "reduced_collector_temperature"
            ]

    _efficiency_plots(
        data=reduced_data,
        electrical_efficiency_labels=sorted(electrical_efficiency_labels),
        logger=logger,
        prefix="ilaria_single_glazed_pc",
        thermal_efficiency_labels=sorted(thermal_efficiency_labels),
    )
    plt.close("all")

    # Plot the portion-covered affect on single-glazed autotherm.
    reduced_data = collections.defaultdict(dict)
    thermal_efficiency_labels = set()
    electrical_efficiency_labels = set()
    for key, sub_dict in data.items():
        if "autotherm_single_glazed_pc_" not in key:
            continue
        portion_covered_match = re.match(PORTION_COVERED_REGEX, key)
        if portion_covered_match is None:
            sys.exit(1)
        portion_covered = float(
            f"{portion_covered_match.group('first_digit')}.{portion_covered_match.group('second_digit')}"
        )
        # Only plot significant portions covered.
        if portion_covered not in {1.0, 0.6, 0.3}:
            continue
        for sub_key, value in sub_dict.items():
            # Fetch the portion covered.
            # Store the specific thermal efficiency.
            reduced_data[sub_key][f"thermal efficiency p.c.={portion_covered}"] = value[
                "thermal_efficiency"
            ]
            thermal_efficiency_labels.add(f"thermal efficiency p.c.={portion_covered}")
            # Store the specific electrical efficiency.
            reduced_data[sub_key][
                f"electrical efficiency p.c.={portion_covered}"
            ] = value["electrical_efficiency"]
            electrical_efficiency_labels.add(
                f"electrical efficiency p.c.={portion_covered}"
            )
            # Store a copy of the reduced temperature
            reduced_data[sub_key]["reduced_collector_temperature"] = value[
                "reduced_collector_temperature"
            ]

    _efficiency_plots(
        data=reduced_data,
        electrical_efficiency_labels=sorted(electrical_efficiency_labels),
        logger=logger,
        prefix="autotherm_single_glazed_pc",
        thermal_efficiency_labels=sorted(thermal_efficiency_labels),
    )
    plt.close("all")

    # Plot the portion-covered affect on double-glazed Ilaria.
    reduced_data = collections.defaultdict(dict)
    thermal_efficiency_labels = set()
    electrical_efficiency_labels = set()
    for key, sub_dict in data.items():
        if "ilaria_double_glazed_pc_" not in key:
            continue
        portion_covered_match = re.match(PORTION_COVERED_REGEX, key)
        if portion_covered_match is None:
            sys.exit(1)
        portion_covered = float(
            f"{portion_covered_match.group('first_digit')}.{portion_covered_match.group('second_digit')}"
        )
        # Only plot significant portions covered.
        if portion_covered not in {1.0, 0.6, 0.3}:
            continue
        for sub_key, value in sub_dict.items():
            # Fetch the portion covered.
            # Store the specific thermal efficiency.
            reduced_data[sub_key][f"thermal efficiency p.c.={portion_covered}"] = value[
                "thermal_efficiency"
            ]
            thermal_efficiency_labels.add(f"thermal efficiency p.c.={portion_covered}")
            # Store the specific electrical efficiency.
            reduced_data[sub_key][
                f"electrical efficiency p.c.={portion_covered}"
            ] = value["electrical_efficiency"]
            electrical_efficiency_labels.add(
                f"electrical efficiency p.c.={portion_covered}"
            )
            # Store a copy of the reduced temperature
            reduced_data[sub_key]["reduced_collector_temperature"] = value[
                "reduced_collector_temperature"
            ]

    _efficiency_plots(
        data=reduced_data,
        electrical_efficiency_labels=sorted(electrical_efficiency_labels),
        logger=logger,
        prefix="ilaria_double_glazed_pc",
        thermal_efficiency_labels=sorted(thermal_efficiency_labels),
    )
    plt.close("all")

    # Plot the portion-covered affect on double-glazed autotherm.
    reduced_data = collections.defaultdict(dict)
    thermal_efficiency_labels = set()
    electrical_efficiency_labels = set()
    for key, sub_dict in data.items():
        if "autotherm_double_glazed_pc_" not in key:
            continue
        portion_covered_match = re.match(PORTION_COVERED_REGEX, key)
        if portion_covered_match is None:
            sys.exit(1)
        portion_covered = float(
            f"{portion_covered_match.group('first_digit')}.{portion_covered_match.group('second_digit')}"
        )
        # Only plot significant portions covered.
        if portion_covered not in {1.0, 0.6, 0.3}:
            continue
        for sub_key, value in sub_dict.items():
            # Fetch the portion covered.
            # Store the specific thermal efficiency.
            reduced_data[sub_key][f"thermal efficiency p.c.={portion_covered}"] = value[
                "thermal_efficiency"
            ]
            thermal_efficiency_labels.add(f"thermal efficiency p.c.={portion_covered}")
            # Store the specific electrical efficiency.
            reduced_data[sub_key][
                f"electrical efficiency p.c.={portion_covered}"
            ] = value["electrical_efficiency"]
            electrical_efficiency_labels.add(
                f"electrical efficiency p.c.={portion_covered}"
            )
            # Store a copy of the reduced temperature
            reduced_data[sub_key]["reduced_collector_temperature"] = value[
                "reduced_collector_temperature"
            ]

    _efficiency_plots(
        data=reduced_data,
        electrical_efficiency_labels=sorted(electrical_efficiency_labels),
        logger=logger,
        prefix="autotherm_double_glazed_pc",
        thermal_efficiency_labels=sorted(thermal_efficiency_labels),
    )
    plt.close("all")

    # Plot the glazing effect on a fully covered Ilaria panel.
    reduced_data = collections.defaultdict(dict)
    thermal_efficiency_labels = set()
    electrical_efficiency_labels = set()
    ilaria_fully_covered_regex = re.compile(r"ilaria_(?P<glazing>[^0-9]*)_pc_1_0_.*")
    for key, sub_dict in data.items():
        ilaria_glazing_match = re.match(ilaria_fully_covered_regex, key)
        if ilaria_glazing_match is None:
            continue
        glazing = ilaria_glazing_match.group("glazing")
        for sub_key, value in sub_dict.items():
            reduced_data[sub_key][f"{glazing} thermal efficiency"] = value[
                "thermal_efficiency"
            ]
            thermal_efficiency_labels.add(f"{glazing} thermal efficiency")
            # Store the specific electrical efficiency.
            reduced_data[sub_key][f"{glazing} electrical efficiency"] = value[
                "electrical_efficiency"
            ]
            electrical_efficiency_labels.add(f"{glazing} electrical efficiency")
            # Store a copy of the reduced temperature
            reduced_data[sub_key]["reduced_collector_temperature"] = value[
                "reduced_collector_temperature"
            ]

    _efficiency_plots(
        data=reduced_data,
        electrical_efficiency_labels=sorted(electrical_efficiency_labels),
        logger=logger,
        prefix="ilaria_pc_1_glazing_comparison",
        thermal_efficiency_labels=sorted(thermal_efficiency_labels),
    )
    plt.close("all")

    # Plot the glazing effect on a fully covered autotherm.
    reduced_data = collections.defaultdict(dict)
    thermal_efficiency_labels = set()
    electrical_efficiency_labels = set()
    autotherm_fully_covered_regex = re.compile(
        r"autotherm_(?P<glazing>[^0-9]*)_pc_1_0_.*"
    )
    for key, sub_dict in data.items():
        autotherm_glazing_match = re.match(autotherm_fully_covered_regex, key)
        if autotherm_glazing_match is None:
            continue
        glazing = autotherm_glazing_match.group("glazing")
        for sub_key, value in sub_dict.items():
            reduced_data[sub_key][f"{glazing} thermal efficiency"] = value[
                "thermal_efficiency"
            ]
            thermal_efficiency_labels.add(f"{glazing} thermal efficiency")
            # Store the specific electrical efficiency.
            reduced_data[sub_key][f"{glazing} electrical efficiency"] = value[
                "electrical_efficiency"
            ]
            electrical_efficiency_labels.add(f"{glazing} electrical efficiency")
            # Store a copy of the reduced temperature
            reduced_data[sub_key]["reduced_collector_temperature"] = value[
                "reduced_collector_temperature"
            ]

    _efficiency_plots(
        data=reduced_data,
        electrical_efficiency_labels=sorted(electrical_efficiency_labels),
        logger=logger,
        prefix="autotherm_pc_1_glazing_comparison",
        thermal_efficiency_labels=sorted(thermal_efficiency_labels),
    )
    plt.close("all")

    # Plot the mass-flow-rate effect on a single-glazed Ilaria.
    reduced_data = collections.defaultdict(dict)
    thermal_efficiency_labels = set()
    electrical_efficiency_labels = set()
    ilaria_mass_flow_rate_regex = re.compile(
        r"ilaria.*single_glazed.*_(?P<first_digit>[0-9]*)_(?P<second_digit>[0-9])_litres_per_hour_.*"
    )
    for key, sub_dict in data.items():
        ilaria_mass_flow_rate_match = re.match(ilaria_mass_flow_rate_regex, key)
        if ilaria_mass_flow_rate_match is None:
            continue
        mass_flow_rate = round(
            int(ilaria_mass_flow_rate_match.group("first_digit"))
            + 0.1 * (int(ilaria_mass_flow_rate_match.group("second_digit"))),
            3,
        )
        if int(mass_flow_rate) == mass_flow_rate:
            mass_flow_rate = int(mass_flow_rate)
        # Only include important mass-flow rates.
        if mass_flow_rate not in [
            round(100, 3),
            round(90, 3),
            round(80, 3),
            round(70, 3),
            round(60, 3),
            round(50, 3),
            round(40, 3),
            round(30, 3),
            round(20, 3),
            round(10, 3),
            round(8, 3),
            round(6, 3),
            round(2, 3),
            round(1, 3),
            round(0.8, 3),
            round(0.6, 3),
            round(0.4, 3),
            round(0.2, 3),
            round(0.1, 3),
        ]:
            continue
        for sub_key, value in sub_dict.items():
            reduced_data[sub_key][f"thermal efficiency {mass_flow_rate}L/h"] = value[
                "thermal_efficiency"
            ]
            thermal_efficiency_labels.add(f"thermal efficiency {mass_flow_rate}L/h")
            # Store the specific electrical efficiency.
            reduced_data[sub_key][f"electrical efficiency {mass_flow_rate}L/h"] = value[
                "electrical_efficiency"
            ]
            electrical_efficiency_labels.add(
                f"electrical efficiency {mass_flow_rate}L/h"
            )
            # Store a copy of the reduced temperature
            reduced_data[sub_key]["reduced_collector_temperature"] = value[
                "reduced_collector_temperature"
            ]

    reduced_data.pop("20.0")

    _efficiency_plots(
        data=reduced_data,
        electrical_efficiency_labels=sorted(electrical_efficiency_labels),
        logger=logger,
        prefix="ilaria_single_glazed_mass_flow_rate_comparison",
        thermal_efficiency_labels=sorted(thermal_efficiency_labels),
    )
    plt.close("all")

    # Plot the mass-flow-rate effect on a single-glazed autotherm.
    reduced_data = collections.defaultdict(dict)
    thermal_efficiency_labels = set()
    electrical_efficiency_labels = set()
    autotherm_mass_flow_rate_regex = re.compile(
        r"autotherm.*single_glazed.*_(?P<first_digit>[0-9]*)_(?P<second_digit>[0-9])_litres_per_hour_.*"
    )
    for key, sub_dict in data.items():
        autotherm_mass_flow_rate_match = re.match(autotherm_mass_flow_rate_regex, key)
        if autotherm_mass_flow_rate_match is None:
            continue
        mass_flow_rate = round(
            int(autotherm_mass_flow_rate_match.group("first_digit"))
            + 0.1 * (int(autotherm_mass_flow_rate_match.group("second_digit"))),
            3,
        )
        if int(mass_flow_rate) == mass_flow_rate:
            mass_flow_rate = int(mass_flow_rate)
        # Only include important mass-flow rates.
        if mass_flow_rate not in [
            round(100, 3),
            round(90, 3),
            round(80, 3),
            round(70, 3),
            round(60, 3),
            round(50, 3),
            round(40, 3),
            round(30, 3),
            round(20, 3),
            round(10, 3),
            round(8, 3),
            round(6, 3),
            round(2, 3),
            round(1, 3),
            round(0.8, 3),
            round(0.6, 3),
            round(0.4, 3),
            round(0.2, 3),
            round(0.1, 3),
        ]:
            continue
        for sub_key, value in sub_dict.items():
            reduced_data[sub_key][f"thermal efficiency {mass_flow_rate}L/h"] = value[
                "thermal_efficiency"
            ]
            thermal_efficiency_labels.add(f"thermal efficiency {mass_flow_rate}L/h")
            # Store the specific electrical efficiency.
            reduced_data[sub_key][f"electrical efficiency {mass_flow_rate}L/h"] = value[
                "electrical_efficiency"
            ]
            electrical_efficiency_labels.add(
                f"electrical efficiency {mass_flow_rate}L/h"
            )
            # Store a copy of the reduced temperature
            reduced_data[sub_key]["reduced_collector_temperature"] = value[
                "reduced_collector_temperature"
            ]

    _efficiency_plots(
        data=reduced_data,
        electrical_efficiency_labels=sorted(electrical_efficiency_labels),
        logger=logger,
        prefix="autotherm_single_glazed_mass_flow_rate_comparison",
        thermal_efficiency_labels=sorted(thermal_efficiency_labels),
    )
    plt.close("all")

    # Plot the mass-flow-rate effect on a double-glazed Ilaria.
    reduced_data = collections.defaultdict(dict)
    thermal_efficiency_labels = set()
    electrical_efficiency_labels = set()
    ilaria_mass_flow_rate_regex = re.compile(
        r"ilaria.*double_glazed.*_(?P<first_digit>[0-9]*)_(?P<second_digit>[0-9])_litres_per_hour_.*"
    )
    for key, sub_dict in data.items():
        ilaria_mass_flow_rate_match = re.match(ilaria_mass_flow_rate_regex, key)
        if ilaria_mass_flow_rate_match is None:
            continue
        mass_flow_rate = round(
            int(ilaria_mass_flow_rate_match.group("first_digit"))
            + 0.1 * (int(ilaria_mass_flow_rate_match.group("second_digit"))),
            3,
        )
        if int(mass_flow_rate) == mass_flow_rate:
            mass_flow_rate = int(mass_flow_rate)
        # Only include important mass-flow rates.
        if mass_flow_rate not in [
            round(100, 3),
            round(90, 3),
            round(80, 3),
            round(70, 3),
            round(60, 3),
            round(50, 3),
            round(40, 3),
            round(30, 3),
            round(20, 3),
            round(10, 3),
            round(8, 3),
            round(6, 3),
            round(2, 3),
            round(1, 3),
            round(0.8, 3),
            round(0.6, 3),
            round(0.4, 3),
            round(0.2, 3),
            round(0.1, 3),
        ]:
            continue
        for sub_key, value in sub_dict.items():
            reduced_data[sub_key][f"thermal efficiency {mass_flow_rate}L/h"] = value[
                "thermal_efficiency"
            ]
            thermal_efficiency_labels.add(f"thermal efficiency {mass_flow_rate}L/h")
            # Store the specific electrical efficiency.
            reduced_data[sub_key][f"electrical efficiency {mass_flow_rate}L/h"] = value[
                "electrical_efficiency"
            ]
            electrical_efficiency_labels.add(
                f"electrical efficiency {mass_flow_rate}L/h"
            )
            # Store a copy of the reduced temperature
            reduced_data[sub_key]["reduced_collector_temperature"] = value[
                "reduced_collector_temperature"
            ]

    _efficiency_plots(
        data=reduced_data,
        electrical_efficiency_labels=sorted(electrical_efficiency_labels),
        logger=logger,
        prefix="ilaria_double_glazed_mass_flow_rate_comparison",
        thermal_efficiency_labels=sorted(thermal_efficiency_labels),
    )
    plt.close("all")

    # Plot the mass-flow-rate effect on a single-glazed autotherm.
    reduced_data = collections.defaultdict(dict)
    thermal_efficiency_labels = set()
    electrical_efficiency_labels = set()
    autotherm_mass_flow_rate_regex = re.compile(
        r"autotherm.*double_glazed.*_(?P<first_digit>[0-9]*)_(?P<second_digit>[0-9])_litres_per_hour_.*"
    )
    for key, sub_dict in data.items():
        autotherm_mass_flow_rate_match = re.match(autotherm_mass_flow_rate_regex, key)
        if autotherm_mass_flow_rate_match is None:
            continue
        mass_flow_rate = round(
            int(autotherm_mass_flow_rate_match.group("first_digit"))
            + 0.1 * (int(autotherm_mass_flow_rate_match.group("second_digit"))),
            3,
        )
        if int(mass_flow_rate) == mass_flow_rate:
            mass_flow_rate = int(mass_flow_rate)
        # Only include important mass-flow rates.
        if mass_flow_rate not in [
            round(100, 3),
            round(90, 3),
            round(80, 3),
            round(70, 3),
            round(60, 3),
            round(50, 3),
            round(40, 3),
            round(30, 3),
            round(20, 3),
            round(10, 3),
            round(8, 3),
            round(6, 3),
            round(2, 3),
            round(1, 3),
            round(0.8, 3),
            round(0.6, 3),
            round(0.4, 3),
            round(0.2, 3),
            round(0.1, 3),
        ]:
            continue
        for sub_key, value in sub_dict.items():
            reduced_data[sub_key][f"thermal efficiency {mass_flow_rate}L/h"] = value[
                "thermal_efficiency"
            ]
            thermal_efficiency_labels.add(f"thermal efficiency {mass_flow_rate}L/h")
            # Store the specific electrical efficiency.
            reduced_data[sub_key][f"electrical efficiency {mass_flow_rate}L/h"] = value[
                "electrical_efficiency"
            ]
            electrical_efficiency_labels.add(
                f"electrical efficiency {mass_flow_rate}L/h"
            )
            # Store a copy of the reduced temperature
            reduced_data[sub_key]["reduced_collector_temperature"] = value[
                "reduced_collector_temperature"
            ]

    _efficiency_plots(
        data=reduced_data,
        electrical_efficiency_labels=sorted(electrical_efficiency_labels),
        logger=logger,
        prefix="autotherm_double_glazed_mass_flow_rate_comparison",
        thermal_efficiency_labels=sorted(thermal_efficiency_labels),
    )

    plt.close("all")

    # Plot the number-of-pipes effect on a single-glazed Ilaria.
    reduced_data = collections.defaultdict(dict)
    thermal_efficiency_labels = set()
    electrical_efficiency_labels = set()
    ilaria_pipes_regex = re.compile(
        r"ilaria_(?P<pipes>[0-9]*)_pipes_(?P<glazing>.*)_pc.*"
    )
    for key, sub_dict in data.items():
        ilaria_pipes_match = re.match(ilaria_pipes_regex, key)
        if ilaria_pipes_match is None:
            continue
        pipes = int(ilaria_pipes_match.group("pipes"))
        glazing = ilaria_pipes_match.group("glazing")
        # if glazing == "unglazed":
        #     continue
        # Only include important mass-flow rates.
        for sub_key, value in sub_dict.items():
            reduced_data[sub_key][
                f"{glazing} thermal efficiency {pipes} pipes"
            ] = value["thermal_efficiency"]
            thermal_efficiency_labels.add(f"{glazing} thermal efficiency {pipes} pipes")
            # Store the specific electrical efficiency.
            reduced_data[sub_key][
                f"{glazing} electrical efficiency {pipes} pipes"
            ] = value["electrical_efficiency"]
            electrical_efficiency_labels.add(
                f"{glazing} electrical efficiency {pipes} pipes"
            )
            # Store a copy of the reduced temperature
            reduced_data[sub_key]["reduced_collector_temperature"] = value[
                "reduced_collector_temperature"
            ]

    _efficiency_plots(
        data=reduced_data,
        electrical_efficiency_labels=sorted(electrical_efficiency_labels),
        logger=logger,
        prefix="ilaria_pc_1_pipes_glazing_comparison",
        thermal_efficiency_labels=sorted(thermal_efficiency_labels),
    )
    plt.close("all")

    # Plot the number-of-pipes effect on a single-glazed Autotherm.
    reduced_data = collections.defaultdict(dict)
    thermal_efficiency_labels = set()
    electrical_efficiency_labels = set()
    autotherm_pipes_regex = re.compile(
        r"autotherm_(?P<pipes>[0-9]*)_pipes_(?P<glazing>.*)_pc.*"
    )
    for key, sub_dict in data.items():
        autotherm_pipes_match = re.match(autotherm_pipes_regex, key)
        if autotherm_pipes_match is None:
            continue
        pipes = int(autotherm_pipes_match.group("pipes"))
        glazing = autotherm_pipes_match.group("glazing")
        if glazing == "unglazed":
            continue
        # Only include important mass-flow rates.
        for sub_key, value in sub_dict.items():
            reduced_data[sub_key][
                f"{glazing} thermal efficiency {pipes} pipes"
            ] = value["thermal_efficiency"]
            thermal_efficiency_labels.add(f"{glazing} thermal efficiency {pipes} pipes")
            # Store the specific electrical efficiency.
            reduced_data[sub_key][
                f"{glazing} electrical efficiency {pipes} pipes"
            ] = value["electrical_efficiency"]
            electrical_efficiency_labels.add(
                f"{glazing} electrical efficiency {pipes} pipes"
            )
            # Store a copy of the reduced temperature
            reduced_data[sub_key]["reduced_collector_temperature"] = value[
                "reduced_collector_temperature"
            ]

    _efficiency_plots(
        data=reduced_data,
        electrical_efficiency_labels=sorted(electrical_efficiency_labels),
        logger=logger,
        prefix="autotherm_pc_1_pipes_glazing_comparison",
        thermal_efficiency_labels=sorted(thermal_efficiency_labels),
    )
    plt.close("all")


def analyse(data_file_directory: str, show_output: Optional[bool] = False) -> None:
    """
    The main method for the analysis module.

    :param data_file_directory:
        The path to the directory containing data files to analyse.

    :param show_output:
        Whether to show the output files generated.

    """

    # * Set up the logger
    logger = get_logger(True, "pvt_analysis", True)

    # * Extract the data.
    data: Dict[str, Dict[str, Any]] = dict()
    for data_file_name in os.listdir(data_file_directory):
        data[data_file_name] = load_model_data(
            os.path.join(data_file_directory, data_file_name)
        )

        # If the data type is not decoupled, steady-state data, then exit.
        data_type = data[data_file_name].pop("data_type")
        if data_type not in (
            STEADY_STATE_DATA_TYPE,
            f"{DECOUPLED_DATA_TYPE}_{STEADY_STATE_DATA_TYPE}",
        ):
            logger.info(
                "Data type was neither 'dynamic' nor 'steady_state'. Omitting..."
            )
            data.pop(data_file_name)
            continue
        if "no_pv" in data_file_name:
            logger.info("Invalid data was removed.")
            data.pop(data_file_name)
            continue

    analyse_decoupled_steady_state_data(data, logger)

    logger.info("Analysis complete - all figures saved successfully.")

    if show_output:
        plt.show()


if __name__ == "__main__":
    parsed_args = _parse_args(sys.argv[1:])
    analyse(parsed_args.data_file_directory, parsed_args.show_output)



#!/usr/bin/python3.7
########################################################################################
# decoupled.py - Module for executing decoupled system runs.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################

"""
The decoupled-run module for the PV-T model.

This module exposes methods needed to carry out a decoupled run of the system.

A decoupled run treates the PV-T collector in isolation, separated from the rest of the
system. Various data about the nature of the panel in this mode of operation is gathered
by calling out to the matrix module.

"""

import datetime
import logging

from typing import Dict, List, Optional, Tuple, Union

import numpy

from dateutil.relativedelta import relativedelta

from .. import convergent_solver, weather
from ..pvt_collector import pvt

from ...__utils__ import BColours, OperatingMode, ProgrammerJudgementFault, SystemData
from ..__utils__ import DivergentSolutionError, time_iterator
from ..constants import CONVERGENT_SOLUTION_PRECISION, DEFAULT_INITIAL_DATE_AND_TIME
from .__utils__ import system_data_from_run

__all__ = (
    "decoupled_dynamic_run",
    "decoupled_steady_state_run",
)


def decoupled_dynamic_run(
    cloud_efficacy_factor: float,
    collector_input_temperature: float,
    days: Optional[int],
    initial_month: int,
    initial_system_temperature_vector: List[float],
    logger: logging.Logger,
    minutes: Optional[int],
    months: Optional[int],
    number_of_pipes: int,
    number_of_temperatures: int,
    number_of_x_elements: int,
    number_of_y_elements: int,
    operating_mode: OperatingMode,
    pvt_collector: pvt.PVT,
    resolution: int,
    save_2d_output: bool,
    start_time: int,
    weather_forecaster: weather.WeatherForecaster,
) -> Tuple[numpy.ndarray, Dict[float, SystemData]]:
    """
    Carries out a dynamic run of the system.

    :param cloud_efficacy_factor:
        How effective the cloud cover is, from 0 (no effect) to 1 (no solar irradiance
        makes it through the cloud layer).

    :param collector_input_temperature:
        The temperature of HTF going into the collector.
        @@@ UNITS UNKNOWN HERE.

    :param days:
        How many days the system should run the simulation for.

    :param initial_month:
        The month for which the simulation should start to be run.

    :param initial_system_temperature_vector:
        The set of system temperatures, expressed as a `list`, at the beginning of the
        run.

    :param logger:
        The logger to be used for the run.

    :param minutes:
        The number of minutes for which to run the simulation, if specified. Otherwise,
        `None`.

    :param months:
        The number of months for which to run the simulation, if specified. Otherwise,
        `None`.

    :param number_of_pipes:
        The number of pipes attached to the absorber.

    :param number_of_temperatures:
        The number of temperatures being modelled.

    :param number_of_x_elements:
        The number of elements in the x direction.

    :param number_of_y_elements:
        The number of elements in the y direction.

    :param operating_mode:
        The operating mode of the system.

    :param pvt_collector:
        A :class:`pvt.PVT` instance representing the PVT panel being modelled.

    :param resolution:
        The temporal resolution (in seconds) for the run.

    :param save_2d_output:
        Whether the 2D output should be saved (True) or not (False).

    :param start_time:
        The start time, expressed in hours from midnight, at which to begin running the
        model.

    :param weather_forecaster:
        The weather forecaster for the system.

    :return:
        A `tuple` containing:
        - the final system-state reached by the model,
        - the system data for the run.

    """

    # Set up a holder for information about the system.
    system_data: Dict[float, SystemData] = dict()

    # Set up the time iterator.
    if minutes is None and days is None and months is None:
        raise ProgrammerJudgementFault(
            "{}Either minutes, days or months must be specified for dynamic ".format(
                BColours.FAIL
            )
            + "runs.{}".format(BColours.ENDC)
        )

    # Determine the number of minutes, days, and months for which to run the simulation.
    num_months = (
        (initial_month if initial_month is not None else 1)
        - 1
        + (months if months is not None else 0)
    )  # [months]
    start_month = (
        initial_month
        if 1 <= initial_month <= 12
        else DEFAULT_INITIAL_DATE_AND_TIME.month
    )  # [months]
    initial_date_and_time = DEFAULT_INITIAL_DATE_AND_TIME.replace(
        hour=start_time, month=start_month
    )

    if minutes is not None:
        final_date_and_time = initial_date_and_time + relativedelta(minutes=minutes)
    elif days is not None:
        final_date_and_time = initial_date_and_time + relativedelta(days=days)
    else:
        final_date_and_time = initial_date_and_time + relativedelta(
            months=num_months % 12, years=num_months // 12
        )

    logger.info(
        "Beginning itterative model:\n  Running from: %s\n  Running to: %s",
        str(initial_date_and_time),
        str(final_date_and_time),
    )

    time_iterator_step = relativedelta(seconds=resolution)

    weather_conditions = weather_forecaster.get_weather(
        pvt_collector.latitude,
        pvt_collector.longitude,
        cloud_efficacy_factor,
        initial_date_and_time,
    )

    system_data[0] = system_data_from_run(
        initial_date_and_time.date(),
        initial_date_and_time,
        number_of_pipes,
        number_of_x_elements,
        operating_mode,
        pvt_collector,
        save_2d_output,
        initial_system_temperature_vector,
        initial_date_and_time.time(),
        weather_conditions,
    )

    previous_run_temperature_vector: Union[
        List[float], numpy.ndarray
    ] = initial_system_temperature_vector

    for run_number, date_and_time in enumerate(
        time_iterator(
            first_time=initial_date_and_time,
            last_time=final_date_and_time,
            resolution=resolution,
            timezone=pvt_collector.timezone,
        )
    ):

        logger.debug(
            "Time: %s: Beginning internal run. Previous temperature vector: T=%s",
            date_and_time.strftime("%d/%m/%Y %H:%M:%S"),
            previous_run_temperature_vector,
        )

        # Determine the "i+1" time.
        next_date_and_time = date_and_time + time_iterator_step

        # Determine the "i+1" current weather conditions.
        try:
            weather_conditions = weather_forecaster.get_weather(
                pvt_collector.latitude,
                pvt_collector.longitude,
                cloud_efficacy_factor,
                next_date_and_time,
            )
        except KeyError as e:
            logger.error("Failed to get weather conditions, do all days have profiles?")
            raise

        try:
            current_run_temperature_vector = (
                convergent_solver.solve_temperature_vector_convergence_method(
                    collector_input_temperature=collector_input_temperature,
                    logger=logger,
                    next_date_and_time=next_date_and_time,
                    number_of_pipes=number_of_pipes,
                    number_of_temperatures=number_of_temperatures,
                    number_of_x_elements=number_of_x_elements,
                    number_of_y_elements=number_of_y_elements,
                    operating_mode=operating_mode,
                    previous_run_temperature_vector=previous_run_temperature_vector,
                    pvt_collector=pvt_collector,
                    resolution=resolution,
                    run_one_temperature_vector=previous_run_temperature_vector,
                    weather_conditions=weather_conditions,
                )
            )
        except DivergentSolutionError as e:
            logger.error(
                "A divergent solution was reached at %s:%s",
                date_and_time.strftime("%D/%M/%Y %H:%M:%S"),
                str(e),
            )
            raise

        # Save the system data output and 2D profiles.
        system_data[run_number + 1] = system_data_from_run(
            next_date_and_time.date(),
            initial_date_and_time,
            number_of_pipes,
            number_of_x_elements,
            operating_mode,
            pvt_collector,
            save_2d_output,
            current_run_temperature_vector,
            next_date_and_time.time(),
            weather_conditions,
        )
        previous_run_temperature_vector = current_run_temperature_vector

    return numpy.asarray(current_run_temperature_vector), system_data


def decoupled_steady_state_run(
    collector_input_temperature: float,
    cloud_efficacy_factor: float,
    initial_date_and_time: datetime.datetime,
    initial_system_temperature_vector: List[float],
    logger: logging.Logger,
    number_of_pipes: int,
    number_of_temperatures: int,
    number_of_x_elements: int,
    number_of_y_elements: int,
    operating_mode: OperatingMode,
    pvt_collector: pvt.PVT,
    save_2d_output: bool,
    weather_forecaster: weather.WeatherForecaster,
) -> Tuple[numpy.ndarray, Dict[float, SystemData]]:
    """
    Carries out a steady-state run of the system.

    :param collector_input_temperature:
        The temperature of HTF going into the collector.
        @@@ UNITS UNKNOWN HERE.

    :param cloud_efficacy_factor:
        How effective the cloud cover is.

    :param initial_date_and_time:
        The date and time for which the system is being run - needed to instantiate the
        weather instance.

    :param initial_system_temperature_vector:
        The initial conditions to use for the system temperature vector.

    :param logger:
        The logger to use for the run.

    :param number_of_pipes:
        The number of pipes attached to the collector.

    :param number_of_temperatures:
        The number of temperatures being modelled in the system.

    :param number_of_x_elements:
        The number of elements in the x direction - i.e., the x-resolution of the run.

    :param number_of_y_elements:
        The number of elements in the y direction - i.e., the y-resolution of the run.

    :param operating_mode:
        The operating mode for the run.

    :param pvt_collector:
        The :class:`pvt.PVT` instance representing the PVT collector being modelled.

    :param save_2d_output:
        Whether 2D output should be saved.

    :param weather_forecaster:
        The weather forecaster to use for the run.

    :return:
        A `tuple` contianing:
        - the final system temperature vector determined at the end of the steady-state
          computation;
        - the system data dictionary about the system once the final conditions have
          been determined.

    """

    # Set up a holder for information about the system.
    system_data: Dict[float, SystemData] = dict()

    # Set up various variables needed to model the system.
    weather_conditions = weather_forecaster.get_weather(
        pvt_collector.latitude,
        pvt_collector.longitude,
        cloud_efficacy_factor,
    )

    logger.info(
        "Beginning steady-state model:\n  Running on: %s\n  Running at convergent "
        "precision: %sK",
        initial_date_and_time.date().strftime("%d/%m/%Y"),
        str(10 ** -CONVERGENT_SOLUTION_PRECISION),
    )

    system_data[0] = system_data_from_run(
        initial_date_and_time.date(),
        initial_date_and_time,
        number_of_pipes,
        number_of_x_elements,
        operating_mode,
        pvt_collector,
        save_2d_output,
        initial_system_temperature_vector,
        initial_date_and_time.time(),
        weather_conditions,
    )

    try:
        current_run_temperature_vector = (
            convergent_solver.solve_temperature_vector_convergence_method(
                collector_input_temperature=collector_input_temperature,
                logger=logger,
                number_of_pipes=number_of_pipes,
                number_of_temperatures=number_of_temperatures,
                number_of_x_elements=number_of_x_elements,
                number_of_y_elements=number_of_y_elements,
                operating_mode=operating_mode,
                pvt_collector=pvt_collector,
                run_one_temperature_vector=initial_system_temperature_vector,
                weather_conditions=weather_conditions,
            )
        )
    except DivergentSolutionError as e:
        logger.error(
            "A divergent solution was when attempting to solve the system in steady-state:%s",
            str(e),
        )
        raise

    # Save the system data output and 2D profiles.
    system_data[1] = system_data_from_run(
        initial_date_and_time.date(),
        initial_date_and_time,
        number_of_pipes,
        number_of_x_elements,
        operating_mode,
        pvt_collector,
        save_2d_output,
        current_run_temperature_vector,
        initial_date_and_time.time(),
        weather_conditions,
    )

    return numpy.asarray(current_run_temperature_vector), system_data



#!/usr/bin/python3.7
########################################################################################
# pvt_collector/glass.py - Represents a glass within a PVT panel.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################

"""
The glass module for the PV-T model.

This module represents a glass layer within a PV-T panel.

"""

from .__utils__ import OpticalLayer

__all__ = ("Glass",)


class Glass(OpticalLayer):
    """
    Represents the glass (upper) layer of the PV-T panel.

    """



#!/usr/bin/python3.7
########################################################################################
# mains_power.py - Represents a mains power source, such as a grid or pipeline network.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2020
########################################################################################

"""
The mains power module for this PV-T system model.

This module represents a mains grid, and contains code and information pertaining to a
mains power grid and gas supply. Among other things, the module is designed with the
goal of computing carbon dioxide outputs generated and mitigating by using the PV-T
system being modelled.

"""

from dataclasses import dataclass
from typing import Any

from ..__utils__ import (
    CarbonEmissions,
    read_yaml,
    TotalPowerData,
)

from .__utils__ import UtilityType


@dataclass
class _Utility:
    """
    Represents a utility, such as electricity or gas.

    .. attribute:: utility_type
        The type of the utility.

    .. attribute:: emissions_per_joule
        The emissions per joule of energy used, measured in kilograms of CO2 equivalent.

    """

    utility_type: UtilityType
    emissions_per_joule: float

    def __repr__(self) -> str:
        """
        Returns a nice-looking representation of the :class:`Utility` instance.

        :return:
            A `str` giving a nice-looking representation of the utility.

        """

        return (
            "Utility("
            f"emissions_per_joule={self.emissions_per_joule}, "
            f"utility_type={self.utility_type}"
            ")"
        )


class MainsSupply:
    """
    Represents a mains power supply.

    """

    # Private Attributes:
    #
    # .. attribute:: _utilities
    #   A mapping of utility type to utility supplied by the mains supply.
    #

    def __init__(self, utilities) -> None:
        """
        Instantiate a mains supply.

        """

        self._utilities = utilities

    def __repr__(self) -> str:
        """
        Returns a nice-looking representation of the mains-power instance.

        :return:
            A `str` giving a representation of the class.

        """

        return f"MainsSupply(utilities={list(self._utilities.values())})"

    @classmethod
    def from_yaml(cls, yaml_data_path: str) -> Any:
        """
        Instantiate a :class:`MainsSupply` instance based on mains supply YAML data.

        :param yaml_data_path:
            The path to the yaml data file containing information about the mains supply

        :return:
            A :class:`MainsSupply` instance based on the YAML data

        """

        yaml_data = read_yaml(yaml_data_path)

        utilities = {
            UtilityType.electricity: _Utility(
                UtilityType.electricity, yaml_data["electricity"]
            ),
            UtilityType.gas: _Utility(UtilityType.gas, yaml_data["natural_gas"]),
        }

        return cls(utilities)

    def get_carbon_emissions(self, total_power_data: TotalPowerData) -> CarbonEmissions:
        """
        Computes the total amount of carbon dioxide produced in the run.

        :param total_power_data:
            Information about the total amount of power produced in the system.

        :return:
            Information about the carbon emissions.

        """

        electrical_carbon_produced = (
            total_power_data.electricity_demand - total_power_data.electricity_supplied
        ) * self._utilities[UtilityType.electricity].emissions_per_joule
        electrical_carbon_saved = (
            total_power_data.electricity_supplied
            * self._utilities[UtilityType.electricity].emissions_per_joule
        )
        heating_carbon_produced = (
            total_power_data.heating_demand - total_power_data.heating_supplied
        ) * self._utilities[UtilityType.gas].emissions_per_joule
        heating_carbon_saved = (
            total_power_data.heating_supplied
            * self._utilities[UtilityType.gas].emissions_per_joule
        )

        return CarbonEmissions(
            electrical_carbon_produced,
            electrical_carbon_saved,
            heating_carbon_produced,
            heating_carbon_saved,
        )



#!/usr/bin/python3.7
########################################################################################
# pvt_collector/eva.py - Represents an eva layer within a PVT panel.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################

"""
The eva module for the PV-T model.

This module represents a eva layer within a PV-T panel.

"""

from .__utils__ import MicroLayer

__all__ = ("EVA",)


class EVA(MicroLayer):
    """
    Represents an EVA layer within the panel.

    """



#!/usr/bin/python3.7
########################################################################################
# pvt_collector/bond.py - Represents an bond layer within a PVT panel.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################

"""
The bond module for the PV-T model.

This module represents a bond layer within a PV-T panel.

"""

from dataclasses import dataclass

from .__utils__ import MicroLayer

__all__ = ("Bond",)


@dataclass
class Bond(MicroLayer):
    """
    Represents an Bond layer within the panel.

    .. attribute:: width
        The width of the bond, measured in meters.

    """

    width: float
